# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data classes for Hoplite databases."""

import dataclasses
import datetime as dt
import enum
from typing import Any, Sequence

from ml_collections import config_dict
import numpy as np


@dataclasses.dataclass(init=False, repr=False, eq=False)
class DynamicInfo:
  """A base dataclass to handle both pre-defined and arbitrary attributes.

  DynamicInfo uses the underlying dataclass implementation to define the set of
  fields that can be set. It also keeps a dictionary of arbitrary attributes
  that are not part of the dataclass. This allows for handling both pre-defined
  and arbitrary attributes in a unified manner.

  Child classes can be defined as:

  ```python
  @dataclasses.dataclass(init=False, repr=False, eq=False)
  class CustomInfo(DynamicInfo):
    required_attr: int
    optional_attr: str | None = None
  ```

  Creating CustomInfo instances is as simple as:

  ```python
  info = CustomInfo(required_attr=1)
  info = CustomInfo(required_attr=1, optional_attr="foo")
  info = CustomInfo(required_attr=1, random_attr="bar")
  info = CustomInfo(required_attr=1, optional_attr="foo", random_attr="bar")
  ```

  Getting and setting attributes, no matter if they are pre-defined or
  arbitrary, can be done via the same interface:

  ```python
  info.required_attr
  info.optional_attr
  info.random_attr
  info.required_attr = 1
  info.optional_attr = "foo"
  info.random_attr = "bar"
  ```
  """

  def __init__(self, **kwargs) -> None:
    # Get the set of fields that are defined in the child dataclass.
    defined_fields = {f.name for f in dataclasses.fields(self)}

    # This will store the arbitrary attributes.
    self._dynamic_info: dict[str, Any] = {}

    # Keep track of fields we haven't seen yet.
    missing_fields = defined_fields.copy()

    # Iterate through all provided keyword arguments.
    for key, value in kwargs.items():
      if key in defined_fields:
        # If the kwarg is a defined field, set it as a normal attribute.
        setattr(self, key, value)
        missing_fields.remove(key)
      else:
        # If it's not a defined field, add it to our dynamic dict.
        self._dynamic_info[key] = value

    # After processing, check if any non-default fields are missing.
    missing_non_default_fields = [
        f.name
        for f in dataclasses.fields(self)
        if (
            f.name in missing_fields
            and f.default is dataclasses.MISSING
            and f.default_factory is dataclasses.MISSING
        )
    ]
    if missing_non_default_fields:
      missing_non_default_fields_str = (
          "'" + "', '".join(missing_non_default_fields) + "'"
      )
      raise TypeError(
          f"'{type(self).__name__}.__init__()' missing required keyword-only"
          f" arguments: {missing_non_default_fields_str}."
      )

    # Manually call `__post_init__` if the user has defined one in a subclass.
    if hasattr(self, "__post_init__"):
      self.__post_init__()

  def __getattr__(self, name: str) -> Any:
    if name in self._dynamic_info:
      return self._dynamic_info[name]
    raise AttributeError(
        f"'{type(self).__name__}' object has no attribute '{name}'."
    )

  def __setattr__(self, name: str, value: Any) -> None:
    # Use the existence of `_dynamic_info` to know if we are initialized.
    defined_fields = {f.name for f in dataclasses.fields(self)}
    if "_dynamic_info" not in self.__dict__ or name in defined_fields:
      super().__setattr__(name, value)
    else:
      self._dynamic_info[name] = value

  def __repr__(self) -> str:
    defined_parts = []
    for f in dataclasses.fields(self):
      value = getattr(self, f.name, dataclasses.MISSING)
      if value is not dataclasses.MISSING:
        defined_parts.append(f"{f.name}={repr(value)}")

    dynamic_parts = [f"{k}={repr(v)}" for k, v in self._dynamic_info.items()]

    all_parts = defined_parts + dynamic_parts
    return f"{self.__class__.__name__}({', '.join(all_parts)})"

  def __eq__(self, other: object) -> bool:
    # Two objects can't be equal if they are not of the same type.
    if self.__class__ is not other.__class__:
      return False

    # Compare both defined fields and `_dynamic_info` dictionaries.
    self_defined_values = tuple(
        getattr(self, f.name) for f in dataclasses.fields(self)
    )
    other_defined_values = tuple(
        getattr(other, f.name) for f in dataclasses.fields(self)
    )
    return (
        self_defined_values == other_defined_values
        and hasattr(self, "_dynamic_info")
        and hasattr(other, "_dynamic_info")
        and self._dynamic_info == other._dynamic_info
    )

  def __getstate__(self) -> dict[str, Any]:
    return {
        "__dict__": self.__dict__.copy(),
        "_dynamic_info": self._dynamic_info.copy(),
    }

  def __setstate__(self, state: dict[str, Any]) -> None:
    self.__dict__.update(state["__dict__"])
    self._dynamic_info.update(state["_dynamic_info"])

  def to_kwargs(self, skip: Sequence[str] | None = None) -> dict[str, Any]:
    """Convert dataclass to a dictionary of keyword arguments.

    Args:
      skip: A sequence of attribute names to skip.

    Returns:
      A dictionary of keyword arguments that can be passed to the constructor to
      create an equivalent object.
    """

    kwargs = {
        f.name: getattr(self, f.name)
        for f in dataclasses.fields(self)
        if getattr(self, f.name) is not dataclasses.MISSING
    }
    if "_dynamic_info" in self.__dict__:
      kwargs.update(self._dynamic_info)
    if skip is not None:
      for key in skip:
        kwargs.pop(key, None)
    return kwargs


@dataclasses.dataclass(init=False, repr=False, eq=False)
class Deployment(DynamicInfo):
  """Deployment (i.e. site) info."""

  id: int
  name: str
  project: str
  latitude: float | None = None
  longitude: float | None = None


@dataclasses.dataclass(init=False, repr=False, eq=False)
class Recording(DynamicInfo):
  """Recording info."""

  id: int
  filename: str
  datetime: dt.datetime | None = None
  deployment_id: int | None = None


@dataclasses.dataclass(init=False, repr=False, eq=False)
class Window(DynamicInfo):
  """Window info."""

  id: int
  recording_id: int
  offsets: list[float]
  embedding: np.ndarray | None


class LabelType(int, enum.Enum):
  NEGATIVE = 0
  POSITIVE = 1
  UNCERTAIN = 2


@dataclasses.dataclass(init=False, repr=False, eq=False)
class Annotation(DynamicInfo):
  """Annotation info."""

  id: int
  recording_id: int
  offsets: list[float]
  label: str
  label_type: LabelType
  provenance: str


@dataclasses.dataclass
class HopliteConfig:
  """Config dataclass used to handle ConfigDict objects in Hoplite databases."""

  def to_config_dict(self) -> config_dict.ConfigDict:
    """Convert to a ConfigDict."""
    return config_dict.ConfigDict(dataclasses.asdict(self))

  @classmethod
  def from_config_dict(cls, config: config_dict.ConfigDict) -> "HopliteConfig":
    """Convert from a ConfigDict."""
    return cls(**config)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Audio source information handling."""

from collections.abc import Iterator
import dataclasses

from etils import epath
from ml_collections import config_dict
from perch_hoplite import audio_io
from perch_hoplite.db import datatypes
import tqdm


@dataclasses.dataclass
class SourceId:
  """Source information for pairing audio with embeddings."""

  dataset_name: str
  file_id: str
  offset_s: float
  shard_len_s: float
  filepath: str
  sample_rate_hz: int

  def to_id(self):
    return f'{self.dataset_name}:{self.file_id}:{self.offset_s}'

  def deployment_name_from_file_id(self) -> str:
    """Returns the deployment name for this source."""
    if '/' in self.file_id:
      return self.file_id.split('/')[0]
    else:
      return self.dataset_name


@dataclasses.dataclass
class AudioSourceConfig(datatypes.HopliteConfig):
  """Configuration for embedding a collection of audio sources.

  Attributes:
    dataset_name: Name of the dataset. (Must be unique for each set of files.)
    base_path: Root directory of the dataset.
    file_glob: Glob pattern for the audio files.
    min_audio_len_s: Minimum audio length to process.
    target_sample_rate_hz: Target sample rate for audio. If -2, use the
      embedding model's declared sample rate. If -1, use the file's native
      sample rate. If > 0, resample to the specified rate.
    shard_len_s: If not None, shard the audio into segments of this length.
    max_shards_per_file: If not None, maximum number of shards per file.
  """

  dataset_name: str
  base_path: str
  file_glob: str
  min_audio_len_s: float = 1.0
  target_sample_rate_hz: int = -2
  shard_len_s: float | None = 60.0
  max_shards_per_file: int | None = None

  def is_compatible(self, other: 'AudioSourceConfig') -> bool:
    """Returns True if other is expected to produce comparable embeddings."""
    return (
        self.dataset_name == other.dataset_name
        and self.target_sample_rate_hz == other.target_sample_rate_hz
        and self.min_audio_len_s == other.min_audio_len_s
    )


@dataclasses.dataclass
class AudioSources(datatypes.HopliteConfig):
  """A collection of AudioSourceConfig, with SourceId iterator."""

  audio_globs: tuple[AudioSourceConfig, ...]
  _file_info_cache: dict[str, tuple[float, int]] = dataclasses.field(
      default_factory=dict, repr=False, hash=False
  )

  def __post_init__(self):
    dataset_names = set(
        audio_glob.dataset_name for audio_glob in self.audio_globs
    )
    if len(dataset_names) < len(self.audio_globs):
      raise ValueError('Dataset names must be unique.')

  def to_config_dict(self) -> config_dict.ConfigDict:
    """Convert to a config dict."""
    globs = tuple(g.to_config_dict() for g in self.audio_globs)
    return config_dict.ConfigDict({'audio_globs': globs})

  @classmethod
  def from_config_dict(cls, config: config_dict.ConfigDict) -> 'AudioSources':
    """Create an AudioSources from a config dict."""
    globs = tuple(
        AudioSourceConfig(**audio_glob) for audio_glob in config.audio_globs
    )
    return cls(audio_globs=globs)

  def merge_update(self, other: 'AudioSources') -> 'AudioSources':
    """Update the audio sources with the new sources.

    Args:
      other: The new audio sources.

    Raises:
      ValueError if any audio globs appear in both and are incompatible.
    Returns:
      A new AudioSources object with the merged audio globs. In case of a
      conflict, the values in the 'other' audio glob takes precedence.
    """
    my_globs = {g.dataset_name: g for g in self.audio_globs}
    other_globs = {g.dataset_name: g for g in other.audio_globs}
    for dataset_name, my_glob in my_globs.items():
      if dataset_name not in other_globs:
        other_globs[dataset_name] = my_glob
      elif not other_globs[dataset_name].is_compatible(my_glob):
        raise ValueError(
            f'Audio glob {other_globs[dataset_name]} '
            f'is incompatible with {my_glob}.'
        )
    return AudioSources(tuple(other_globs.values()))

  def _get_audio_len_s_and_sample_rate_hz(
      self, filepath: epath.Path
  ) -> tuple[float, int]:
    """Returns the audio length and sample rate of the audio file."""
    filepath_posix = filepath.as_posix()
    if filepath_posix in self._file_info_cache:
      return self._file_info_cache[filepath_posix]
    audio_len_s, sample_rate_hz = audio_io.get_file_length_s_and_sample_rate(
        filepath_posix
    )
    self._file_info_cache[filepath_posix] = (audio_len_s, sample_rate_hz)
    return audio_len_s, sample_rate_hz

  def iterate_all_sources(
      self,
      target_dataset_name: str | None = None,
  ) -> Iterator[SourceId]:
    """Yields all sources for all datasets (or just a single dataset).

    Args:
      target_dataset_name: If not None, only yield sources for this dataset.

    Yields:
      SourceId objects.
    """
    for glob in self.audio_globs:
      if (
          target_dataset_name is not None
          and glob.dataset_name != target_dataset_name
      ):
        continue
      # If base_path is a URL, the posix path may not match the original string.
      base_path = epath.Path(glob.base_path)
      filepaths = tuple(base_path.glob(glob.file_glob))
      shard_len_s = glob.shard_len_s
      max_shards_per_file = glob.max_shards_per_file

      for filepath in tqdm.tqdm(filepaths):
        file_id = filepath.as_posix()[len(base_path.as_posix()) + 1 :]
        audio_len_s, sample_rate_hz = self._get_audio_len_s_and_sample_rate_hz(
            filepath
        )
        if shard_len_s is None:
          yield SourceId(
              dataset_name=glob.dataset_name,
              file_id=file_id,
              offset_s=0,
              shard_len_s=-1,
              filepath=filepath.as_posix(),
              sample_rate_hz=sample_rate_hz,
          )
          continue

        # Otherwise, need to emit sharded SourceId's.
        if audio_len_s <= 0:
          continue
        shard_num = 0
        while max_shards_per_file is None or shard_num < max_shards_per_file:
          offset_s = shard_num * shard_len_s
          if offset_s >= audio_len_s:
            break
          # When the new shard extends beyond the end of the audio, and the
          # shard will be shorter than the minimum audio length, we are done.
          if (
              offset_s + shard_len_s > audio_len_s
              and audio_len_s - offset_s < glob.min_audio_len_s
          ):
            break
          yield SourceId(
              dataset_name=glob.dataset_name,
              file_id=file_id,
              offset_s=offset_s,
              shard_len_s=shard_len_s,
              filepath=filepath.as_posix(),
              sample_rate_hz=sample_rate_hz,
          )
          shard_num += 1



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Hoplite search."""

import shutil
import tempfile

import numpy as np
from perch_hoplite.db.tests import test_utils

from absl.testing import absltest
from absl.testing import parameterized

EMBEDDING_SIZE = 128


class SearchTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  @parameterized.parameters(*test_utils.DB_TYPES)
  def test_exact_search(self, db_type):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 100, rng, EMBEDDING_SIZE)

    # Pick a random window to search for
    window_ids = db.match_window_ids()
    target_id = window_ids[0]
    target_embedding = db.get_embedding(target_id)

    # Search for the target embedding exactly
    results = db.search(
        target_embedding,
        search_list_size=5,
        approximate=False,
        score_fn_name='cos',
    )

    # We expect the target_id to be one of the top results (score ~1 for cosine)
    found_ids = [r.window_id for r in results]
    self.assertIn(target_id, found_ids)

    # Since we used the exact embedding, the score for target_id should be ~1.0
    for result in results:
      if result.window_id == target_id:
        self.assertAlmostEqual(result.sort_score, 1.0, places=4)

  def test_sqlite_approximate_search(self):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(
        self.tempdir, 'sqlite_usearch', 100, rng, EMBEDDING_SIZE
    )

    # Pick a random window to search for
    window_ids = db.match_window_ids()
    target_id = window_ids[0]
    target_embedding = db.get_embedding(target_id)

    # Search approximately
    results = db.search(target_embedding, search_list_size=5, approximate=True)

    # In such a small DB, approximate search should find the exact match easily
    found_ids = [r.window_id for r in results]
    self.assertIn(target_id, found_ids)

    for r in results:
      if r.window_id == target_id:
        # The score should be high.
        # Whether it's exactly 1.0 depends on normalization.
        self.assertGreater(r.sort_score, 0.1)

  def test_sqlite_approximate_target_score_error(self):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(
        self.tempdir, 'sqlite_usearch', 10, rng, EMBEDDING_SIZE
    )
    query_embedding = rng.normal(size=(EMBEDDING_SIZE,))
    with self.assertRaisesRegex(
        ValueError, 'Approximate search does not support target_score'
    ):
      db.search(
          query_embedding,
          search_list_size=5,
          approximate=True,
          target_score=0.5,
      )

  @parameterized.parameters(*test_utils.DB_TYPES)
  def test_target_score_search(self, db_type):
    rng = np.random.default_rng(42)
    db = test_utils.make_db(self.tempdir, db_type, 100, rng, EMBEDDING_SIZE)

    query_embedding = rng.normal(size=(EMBEDDING_SIZE,))
    # Search for embeddings with score near 0.5
    target_score = 0.5
    results = db.search(
        query_embedding,
        search_list_size=5,
        approximate=False,
        target_score=target_score,
        score_fn_name='cos',
    )
    # The scores should be the actual cosine similarity, because of reranking.
    # So they should be in the range [-1, 1], and likely close to 0.5 if
    # the database is large enough.
    for result in results:
      self.assertGreaterEqual(result.sort_score, -1.0)
      self.assertLessEqual(result.sort_score, 1.0)
      # Since we targeted 0.5, the absolute difference should be small-ish.
      self.assertLess(abs(result.sort_score - target_score), 0.5)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Database of bioacoustic label domains."""

import dataclasses
import functools
import json
import os
import typing

from perch_hoplite import path_utils
from perch_hoplite.taxonomy import namespace

TAXONOMY_DATABASE_FILENAME = "taxonomy/taxonomy_database.json"


ClassListType = str | namespace.ClassList | tuple[str, ...]
MappingType = str | namespace.Mapping | dict[str, str]


def get_classes(class_list: ClassListType) -> tuple[str, ...]:
  """Load classes from the namespace database.

  Args:
    class_list: Name of the class list to load. This can be a class list,
      namespace, or mapping name. If it is a mapping name, the sorted tuple of
      all target classes is returned. If an actual ClassList is passed, the
      tuple of classes is returned.

  Returns:
    A tuple of classes.
  """
  if isinstance(class_list, namespace.ClassList):
    return class_list.classes
  elif isinstance(class_list, tuple):
    return class_list

  db = load_db()
  if class_list in db.class_lists:
    return db.class_lists[class_list].classes
  elif class_list in db.namespaces:
    return tuple(sorted(tuple(db.namespaces[class_list].classes)))
  elif class_list in db.mappings:
    image_classes = db.mappings[class_list].mapped_pairs.values()
    return tuple(sorted(tuple(image_classes)))
  else:
    raise ValueError(
        "Class list %s not found in namespace database." % class_list
    )


def get_mapping(mapping: MappingType) -> dict[str, str]:
  """Load mapping from the namespace database."""
  if isinstance(mapping, namespace.Mapping):
    return mapping.mapped_pairs
  elif isinstance(mapping, dict):
    return mapping
  db = load_db()
  if mapping in db.mappings:
    return db.mappings[mapping].mapped_pairs
  else:
    raise ValueError("Mapping %s not found in namespace database." % mapping)


def num_classes(class_list: ClassListType) -> int:
  """Return the number of classes in the class list."""
  return len(get_classes(class_list))


@dataclasses.dataclass
class TaxonomyDatabase:
  namespaces: dict[str, namespace.Namespace]
  class_lists: dict[str, namespace.ClassList]
  mappings: dict[str, namespace.Mapping]


def validate_taxonomy_database(taxonomy_database: TaxonomyDatabase) -> None:
  """Validate the taxonomy database.

  This ensures that all class lists, namespaces, and mappings are consistent.

  Args:
    taxonomy_database: A taxonomy database structure to validate.

  Raises:
    ValueError or KeyError when the database is invalid.
  """
  namespaces = taxonomy_database.namespaces

  for mapping_name, mapping in taxonomy_database.mappings.items():
    if (
        set(mapping.mapped_pairs.keys())
        - namespaces[mapping.source_namespace].classes
    ):
      raise ValueError(
          f"Mapping {mapping_name} contains a source class not in "
          f"the namespace ({mapping.source_namespace})."
      )
    if (
        set(mapping.mapped_pairs.values())
        - namespaces[mapping.target_namespace].classes
    ):
      raise ValueError(
          f"Mapping {mapping_name} contains a target class not in "
          f"the namespace ({mapping.target_namespace})."
      )

  for class_name, class_list in taxonomy_database.class_lists.items():
    classes = class_list.classes
    if (
        set(classes)
        - namespaces[class_list.namespace].classes
        - {namespace.UNKNOWN_LABEL}
    ):
      raise ValueError(
          f"ClassList {class_name} contains a class not in "
          f"the namespace ({class_list.namespace})."
      )


def load_taxonomy_database(
    taxonomy_database: dict[str, typing.Any],
) -> TaxonomyDatabase:
  """Construct a taxonomy database from a dictionary.

  Args:
    taxonomy_database: The database as loaded from a JSON file.

  Returns:
    A taxonomy database.

  Raises:
    TypeError when the database contains unknown keys.
  """
  namespaces = {
      name: namespace.Namespace(
          classes=frozenset(namespace_.pop("classes")), **namespace_
      )
      for name, namespace_ in taxonomy_database.pop("namespaces").items()
  }
  class_lists = {
      name: namespace.ClassList(
          classes=tuple(class_list.pop("classes")), **class_list
      )
      for name, class_list in taxonomy_database.pop("class_lists").items()
  }
  mappings = {
      name: namespace.Mapping(**mapping)
      for name, mapping in taxonomy_database.pop("mappings").items()
  }
  return TaxonomyDatabase(
      namespaces=namespaces,
      class_lists=class_lists,
      mappings=mappings,
      **taxonomy_database,
  )


class TaxonomyDatabaseEncoder(json.JSONEncoder):

  def default(self, o):
    if isinstance(o, frozenset):
      return sorted(o)
    return super().default(o)


def dump_db(taxonomy_database: TaxonomyDatabase, validate: bool = True) -> str:
  if validate:
    validate_taxonomy_database(taxonomy_database)
  return json.dumps(
      dataclasses.asdict(taxonomy_database),
      cls=TaxonomyDatabaseEncoder,
      indent=2,
      sort_keys=True,
  )


@functools.cache
def load_db(
    path: os.PathLike[str] | str = TAXONOMY_DATABASE_FILENAME,
    validate: bool = True,
) -> TaxonomyDatabase:
  """Load the taxonomy database.

  This loads the taxonomy database from the given JSON file. It converts the
  database into Python data structures and optionally validates that the
  database is consistent.

  Args:
    path: The JSON file to load.
    validate: If true, it validates the database.

  Returns:
    The taxonomy database.
  """
  with path_utils.open_file(path, "r") as f:
    data = json.load(f)
  taxonomy_database = load_taxonomy_database(data)
  if validate:
    validate_taxonomy_database(taxonomy_database)
  return taxonomy_database



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for metadata handling."""

import os
import shutil
import tempfile

from perch_hoplite.agile import metadata
from perch_hoplite.db import datatypes

from absl.testing import absltest


class MetadataTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()
    self.desc_path = os.path.join(
        self.tempdir, 'hoplite_metadata_description.csv'
    )
    self.dep_path = os.path.join(
        self.tempdir, 'hoplite_deployments_metadata.csv'
    )
    self.rec_path = os.path.join(
        self.tempdir, 'hoplite_recordings_metadata.csv'
    )

    with open(self.desc_path, 'w') as f:
      f.write(
          'field_name,metadata_level,type,description\n'
          'deployment,deployment,str,Deployment name\n'
          'habitat,deployment,str,Habitat type\n'
          'lon,deployment,float,"Longitude, degrees"\n'
          'lat,deployment,float,"Latitude, degrees"\n'
          'recorder_id,deployment,bytes,ID of hardware\n'
          'recording,recording,str,Recording name\n'
          'observer,recording,str,Observer name\n'
          'temp_c,recording,float,"Temperature, C"\n'
      )
    with open(self.dep_path, 'w') as f:
      f.write(
          'deployment,habitat,lon,lat,recorder_id\n'
          'dep_a,"forest",-120.1,30.2,rec_01\n'
          'dep_b,"prairie",-120.2,30.2,rec_02\n'
      )
    with open(self.rec_path, 'w') as f:
      f.write('recording,observer,temp_c\nrec_a,Buffy,\nrec_b,Willow,22.2\n')

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  def test_metadata_loading(self):
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    self.assertLen(md.deployment_metadata, 2)
    self.assertLen(md.recording_metadata, 2)

  def test_get_deployment_metadata(self):
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    print(md.recording_metadata)
    dep_a_md = md.get_deployment_metadata('dep_a')
    self.assertEqual(
        dep_a_md,
        {
            'deployment': 'dep_a',
            'habitat': 'forest',
            'lon': -120.1,
            'lat': 30.2,
            'recorder_id': b'rec_01',
        },
    )

  def test_get_recording_metadata(self):
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    rec_b_md = md.get_recording_metadata('rec_b')
    self.assertEqual(
        rec_b_md,
        {'recording': 'rec_b', 'observer': 'Willow', 'temp_c': 22.2},
    )

  def test_get_recording_metadata_missing_value(self):
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    rec_a_md = md.get_recording_metadata('rec_a')
    self.assertEqual(
        rec_a_md,
        {'recording': 'rec_a', 'observer': 'Buffy', 'temp_c': None},
    )

  def test_get_unknown_metadata(self):
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    self.assertEqual({}, md.get_deployment_metadata('unknown'))
    self.assertEqual({}, md.get_recording_metadata('unknown'))

  def test_missing_description_file(self):
    os.remove(self.desc_path)
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    self.assertEqual({}, md.get_deployment_metadata('dep_a'))
    self.assertEqual({}, md.get_recording_metadata('rec_a'))

  def test_missing_deployment_file(self):
    os.remove(self.dep_path)
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    self.assertEqual({}, md.get_deployment_metadata('dep_a'))
    self.assertLen(md.recording_metadata, 2)

  def test_missing_recording_file(self):
    os.remove(self.rec_path)
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    self.assertLen(md.deployment_metadata, 2)
    self.assertEqual({}, md.get_recording_metadata('rec_a'))

  def test_get_annotations(self):
    ann_path = os.path.join(self.tempdir, 'hoplite_annotations.csv')
    with open(ann_path, 'w') as f:
      f.write(
          'recording,label,start_offset_s,end_offset_s,label_type\n'
          'rec_a,species_a,0.5,1.5,positive\n'
          'rec_a,species_b,2.0,3.0,positive\n'
          'rec_b,species_a,1.0,2.0,negative\n'
      )
    md = metadata.AgileMetadata.from_directory(self.tempdir)
    rec_a_ann = md.get_recording_annotations('rec_a')
    self.assertEqual(
        rec_a_ann,
        [
            datatypes.Annotation(
                id=-1,
                recording_id=-1,
                label='species_a',
                offsets=[0.5, 1.5],
                label_type=datatypes.LabelType.POSITIVE,
                provenance='agile_metadata',
            ),
            datatypes.Annotation(
                id=-1,
                recording_id=-1,
                label='species_b',
                offsets=[2.0, 3.0],
                label_type=datatypes.LabelType.POSITIVE,
                provenance='agile_metadata',
            ),
        ],
    )
    rec_b_ann = md.get_recording_annotations('rec_b')
    self.assertEqual(
        rec_b_ann,
        [
            datatypes.Annotation(
                id=-1,
                recording_id=-1,
                label='species_a',
                offsets=[1.0, 2.0],
                label_type=datatypes.LabelType.NEGATIVE,
                provenance='agile_metadata',
            )
        ],
    )
    rec_c_ann = md.get_recording_annotations('rec_c')
    self.assertEqual(rec_c_ann, [])

    with self.subTest('no annotations'):
      os.remove(ann_path)
      md = metadata.AgileMetadata.from_directory(self.tempdir)
      self.assertEqual([], md.get_recording_annotations('rec_a'))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for classifier data management."""

import shutil
import tempfile

from ml_collections import config_dict
import numpy as np
from perch_hoplite.agile import classifier_data
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface
from perch_hoplite.db.tests import test_utils as db_test_utils

from absl.testing import absltest


class ClassifierDataTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  def make_db_with_labels(
      self,
      num_embeddings: int,
      unlabeled_prob: float,
      positive_label_prob: float,
      rng: np.random.Generator,
  ) -> interface.HopliteDBInterface:
    db = db_test_utils.make_db(
        self.tempdir,
        'in_mem',
        num_embeddings=num_embeddings,
        rng=rng,
    )
    db_test_utils.add_random_labels(
        db,
        rng=rng,
        unlabeled_prob=unlabeled_prob,
        positive_label_prob=positive_label_prob,
    )
    return db

  def test_train_test_split_fully_labeled(self):
    num_embeddings = 5_000
    db = self.make_db_with_labels(
        num_embeddings=num_embeddings,
        unlabeled_prob=0.0,
        positive_label_prob=0.5,
        rng=np.random.default_rng(42),
    )
    data_manager = classifier_data.AgileDataManager(
        target_labels=db_test_utils.CLASS_LABELS,
        db=db,
        train_ratio=0.8,
        min_eval_examples=10,
        batch_size=10,
        weak_negatives_batch_size=10,
        rng=np.random.default_rng(42),
    )
    train_ids, eval_ids = data_manager.get_train_test_split()
    # Because classes can overlap, we can't guarantee that the train and eval
    # sets are exactly 80% and 20%, but we can check that they are close.
    self.assertGreater(len(eval_ids), 0.9 * 0.2 * num_embeddings)
    self.assertLess(len(eval_ids), 1.1 * 0.2 * num_embeddings)
    self.assertLen(train_ids, num_embeddings - len(eval_ids))
    # But we really want to check that the two sets are disjoint.
    self.assertEmpty(np.intersect1d(train_ids, eval_ids))

    # Check that the labeled examples are split correctly.
    for label in db_test_utils.CLASS_LABELS:
      class_label_ids = db.match_window_ids(
          annotations_filter=config_dict.create(eq=dict(label=label))
      )
      eval_count = np.intersect1d(class_label_ids, eval_ids).shape[0]
      self.assertGreaterEqual(eval_count, 10)

  def test_train_test_split_partially_labeled(self):
    num_embeddings = 5_000
    unlabeled_prob = 0.5
    db = self.make_db_with_labels(
        num_embeddings=num_embeddings,
        unlabeled_prob=unlabeled_prob,
        positive_label_prob=0.5,
        rng=np.random.default_rng(42),
    )
    data_manager = classifier_data.AgileDataManager(
        target_labels=db_test_utils.CLASS_LABELS,
        db=db,
        train_ratio=0.8,
        min_eval_examples=10,
        batch_size=10,
        weak_negatives_batch_size=10,
        rng=np.random.default_rng(42),
    )
    train_ids, eval_ids = data_manager.get_train_test_split()
    # Because classes can overlap, we can't guarantee that the train and eval
    # sets are exactly 80% and 20%, but we can check that they are close.
    self.assertGreater(
        len(eval_ids), unlabeled_prob * 0.9 * 0.2 * num_embeddings
    )
    self.assertLess(len(eval_ids), unlabeled_prob * 1.1 * 0.2 * num_embeddings)
    self.assertGreater(
        len(train_ids), unlabeled_prob * 0.9 * 0.8 * num_embeddings
    )
    self.assertLess(len(train_ids), unlabeled_prob * 1.1 * 0.8 * num_embeddings)
    # But we really want to check that the two sets are disjoint.
    self.assertEmpty(np.intersect1d(train_ids, eval_ids))

  def test_partial_classes(self):
    num_embeddings = 5_000
    db = self.make_db_with_labels(
        num_embeddings=num_embeddings,
        # Add a label to every embedding.
        unlabeled_prob=0.0,
        positive_label_prob=0.5,
        rng=np.random.default_rng(42),
    )
    # Only use three labels, which is half the length of the full class list.
    data_manager = classifier_data.AgileDataManager(
        target_labels=db_test_utils.CLASS_LABELS[:3],
        db=db,
        train_ratio=0.8,
        min_eval_examples=10,
        batch_size=10,
        weak_negatives_batch_size=10,
        rng=np.random.default_rng(42),
    )
    train_ids, eval_ids = data_manager.get_train_test_split()
    self.assertEmpty(np.intersect1d(train_ids, eval_ids))
    for label in db_test_utils.CLASS_LABELS[:3]:
      class_label_ids = db.match_window_ids(
          annotations_filter=config_dict.create(eq=dict(label=label))
      )
      eval_count = np.intersect1d(class_label_ids, eval_ids).shape[0]
      self.assertGreaterEqual(eval_count, 10)
    # Show that examples without target labels are not included.
    for label in db_test_utils.CLASS_LABELS[3:]:
      class_label_ids = db.match_window_ids(
          annotations_filter=config_dict.create(eq=dict(label=label))
      )
      eval_count = np.intersect1d(class_label_ids, eval_ids).shape[0]
      train_count = np.intersect1d(class_label_ids, train_ids).shape[0]
      self.assertEqual(eval_count, 0)
      self.assertEqual(train_count, 0)

  def test_auto_labels(self):
    num_embeddings = 5_000
    db = self.make_db_with_labels(
        num_embeddings=num_embeddings,
        # Add a label to every embedding.
        unlabeled_prob=0.0,
        positive_label_prob=0.5,
        rng=np.random.default_rng(42),
    )
    # Only use three labels, which is half the length of the full class list.
    data_manager = classifier_data.AgileDataManager(
        target_labels=None,
        db=db,
        train_ratio=0.8,
        min_eval_examples=10,
        batch_size=10,
        weak_negatives_batch_size=10,
        rng=np.random.default_rng(42),
    )
    self.assertLen(
        data_manager.get_target_labels(), len(db_test_utils.CLASS_LABELS)
    )

  def test_multihot_labels(self):
    db = db_test_utils.make_db(
        self.tempdir,
        'in_mem',
        num_embeddings=100,
        rng=np.random.default_rng(42),
    )
    data_manager = classifier_data.AgileDataManager(
        target_labels=db_test_utils.CLASS_LABELS,
        db=db,
        train_ratio=0.8,
        min_eval_examples=10,
        batch_size=10,
        weak_negatives_batch_size=10,
        rng=np.random.default_rng(42),
    )

    def add_label(window_id, lbl_idx, lbl_type):
      window = db.get_window(window_id)
      db.insert_annotation(
          recording_id=window.recording_id,
          offsets=window.offsets,
          label=db_test_utils.CLASS_LABELS[lbl_idx],
          label_type=lbl_type,
          provenance='test',
          handle_duplicates='allow',
      )

    with self.subTest('single_positive_label'):
      add_label(1, 3, datatypes.LabelType.POSITIVE)
      multihot, mask = data_manager.get_multihot_labels(1)
      np.testing.assert_equal(multihot, (0, 0, 0, 1, 0, 0))
      np.testing.assert_equal(mask, (0, 0, 0, 1, 0, 0))

    with self.subTest('single_negative_label'):
      add_label(2, 3, datatypes.LabelType.NEGATIVE)
      multihot, mask = data_manager.get_multihot_labels(2)
      np.testing.assert_equal(multihot, (0, 0, 0, 0, 0, 0))
      np.testing.assert_equal(mask, (0, 0, 0, 1, 0, 0))

    with self.subTest('disagreeing_labels'):
      add_label(3, 1, datatypes.LabelType.POSITIVE)
      add_label(3, 1, datatypes.LabelType.POSITIVE)
      add_label(3, 1, datatypes.LabelType.POSITIVE)
      add_label(3, 1, datatypes.LabelType.NEGATIVE)
      multihot, mask = data_manager.get_multihot_labels(3)
      np.testing.assert_equal(multihot, (0, 0.75, 0, 0, 0, 0))
      np.testing.assert_equal(mask, (0, 1, 0, 0, 0, 0))

    with self.subTest('multiple_labels'):
      add_label(4, 1, datatypes.LabelType.POSITIVE)
      add_label(4, 1, datatypes.LabelType.POSITIVE)
      add_label(4, 2, datatypes.LabelType.POSITIVE)
      add_label(4, 3, datatypes.LabelType.NEGATIVE)
      add_label(4, 5, datatypes.LabelType.POSITIVE)
      multihot, mask = data_manager.get_multihot_labels(4)
      np.testing.assert_equal(multihot, (0, 1, 1, 0, 0, 1))
      np.testing.assert_equal(mask, (0, 1, 1, 1, 0, 1))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metrics for training and validation."""

from typing import Any, Dict

import numpy as np


def map_(
    logits: np.ndarray,
    labels: np.ndarray,
    label_mask: np.ndarray | None = None,
    sort_descending: bool = True,
) -> np.ndarray:
  return average_precision(
      scores=logits,
      labels=labels,
      label_mask=label_mask,
      sort_descending=sort_descending,
  )


def cmap(
    logits: np.ndarray,
    labels: np.ndarray,
    sort_descending: bool = True,
    sample_threshold: int = 0,
) -> Dict[str, Any]:
  """Class mean average precision."""
  class_aps = average_precision(
      scores=logits.T, labels=labels.T, sort_descending=sort_descending
  )
  mask = np.sum(labels, axis=0) > sample_threshold
  class_aps = np.where(mask, class_aps, np.nan)
  macro_cmap = np.mean(class_aps, where=mask)
  return {
      'macro': macro_cmap,
      'individual': class_aps,
  }


def roc_auc(
    logits: np.ndarray,
    labels: np.ndarray,
    label_mask: np.ndarray | None = None,
    sort_descending: bool = True,
    sample_threshold: int = 1,
) -> Dict[str, Any]:
  """Computes ROC-AUC scores.

  Args:
    logits: A score for each label which can be ranked.
    labels: A multi-hot encoding of the ground truth positives. Must match the
      shape of scores.
    label_mask: A mask indicating which labels to involve in the calculation.
    sort_descending: An indicator if the search result ordering is in descending
      order (e.g. for evaluating over similarity metrics where higher scores are
      preferred). If false, computes average_precision on descendingly sorted
      inputs.
    sample_threshold: Only classes with at least this many samples will be used
      in the calculation of the final metric. By default this is 1, which means
      that classes without any positive examples will be ignored.

  Returns:
    A dictionary of ROC-AUC scores using the arithmetic ('macro') and
    geometric means, along with individual class ('individual') ROC-AUC and its
    variance.
  """
  if label_mask is not None:
    label_mask = label_mask.T
  class_roc_auc, class_roc_auc_var = generalized_mean_rank(
      logits.T, labels.T, label_mask=label_mask, sort_descending=sort_descending
  )
  mask = np.sum(labels, axis=0) >= sample_threshold
  class_roc_auc = np.where(mask, class_roc_auc, np.nan)
  class_roc_auc_var = np.where(mask, class_roc_auc_var, np.nan)
  return {
      'macro': np.mean(class_roc_auc, where=mask),
      'geometric': np.exp(np.mean(np.log(class_roc_auc), where=mask)),
      'individual': class_roc_auc,
      'individual_var': class_roc_auc_var,
  }


def average_precision(
    scores: np.ndarray,
    labels: np.ndarray,
    label_mask: np.ndarray | None = None,
    sort_descending: bool = True,
    interpolated: bool = False,
) -> np.ndarray:
  """Average precision.

  The average precision is the area under the precision-recall curve. When
  using interpolation we take the maximum precision over all smaller recalls.
  The intuition is that it often makes sense to evaluate more documents if the
  total percentage of relevant documents increases.
  Average precision is computed over the last axis.

  Args:
    scores: A score for each label which can be ranked.
    labels: A multi-hot encoding of the ground truth positives. Must match the
      shape of scores.
    label_mask: A mask indicating which labels to involve in the calculation.
    sort_descending: An indicator if the search result ordering is in descending
      order (e.g. for evaluating over similarity metrics where higher scores are
      preferred). If false, computes average_precision on descendingly sorted
      inputs.
    interpolated: Whether to use interpolation.

  Returns:
    The average precision.
  """
  if label_mask is not None:
    # Set all masked labels to zero, and send the scores for those labels to a
    # low/high value (depending on whether we sort in descending order or not).
    # Then the masked scores+labels will not impact the average precision
    # calculation.
    labels = labels * label_mask
    extremum_score = (
        np.min(scores) - 1.0 if sort_descending else np.max(scores) + 1.0
    )
    scores = np.where(label_mask, scores, extremum_score)
  idx = np.argsort(scores)
  if sort_descending:
    idx = np.flip(idx, axis=-1)
  scores = np.take_along_axis(scores, idx, axis=-1)
  labels = np.take_along_axis(labels, idx, axis=-1)
  pr_curve = np.cumsum(labels, axis=-1) / (np.arange(labels.shape[-1]) + 1)
  if interpolated:
    pr_curve = -np.maximum.accumulate(-pr_curve, axis=-1)

  # In case of an empty row, assign precision = 0, and avoid dividing by zero.
  mask = np.float32(np.sum(labels, axis=-1) != 0)
  raw_av_prec = np.sum(pr_curve * labels, axis=-1) / np.maximum(
      np.sum(labels, axis=-1), 1.0
  )
  return mask * raw_av_prec


def generalized_mean_rank(
    scores: np.ndarray,
    labels: np.ndarray,
    label_mask: np.ndarray | None = None,
    sort_descending: bool = True,
) -> tuple[np.ndarray, np.ndarray]:
  """Computes the generalized mean rank and its variance over the last axis.

  The generalized mean rank can be expressed as

      (sum_i #P ranked above N_i) / (#P * #N),

  or equivalently,

      1 - (sum_i #N ranked above P_i) / (#P * #N).

  This metric is usually better visualized in the logits domain, where it
  reflects the log-odds of ranking a randomly-chosen positive higher than a
  randomly-chosen negative.

  Args:
    scores: A score for each label which can be ranked.
    labels: A multi-hot encoding of the ground truth positives. Must match the
      shape of scores.
    label_mask: A mask indicating which labels to involve in the calculation.
    sort_descending: An indicator if the search result ordering is in descending
      order (e.g. for evaluating over similarity metrics where higher scores are
      preferred). If false, computes the generalize mean rank on descendingly
      sorted inputs.

  Returns:
    The generalized mean rank and its variance. The variance is calculated by
    considering each positive to be an independent sample of the value
    1 - #N ranked above P_i / #N. This gives a measure of how consistently
    positives are ranked.
  """
  idx = np.argsort(scores, axis=-1)
  if sort_descending:
    idx = np.flip(idx, axis=-1)
  labels = np.take_along_axis(labels, idx, axis=-1)
  if label_mask is None:
    label_mask = True
  else:
    label_mask = np.take_along_axis(label_mask, idx, axis=-1)

  num_p = (labels > 0).sum(axis=-1, where=label_mask)
  # Prevent divsion by zero is num_p is zero.
  num_p = np.maximum(num_p, 1)
  num_p_above = np.cumsum((labels > 0) & label_mask, axis=-1)
  num_n = (labels == 0).sum(axis=-1, where=label_mask)
  num_n_above = np.cumsum((labels == 0) & label_mask, axis=-1)

  gmr = num_p_above.mean(axis=-1, where=(labels == 0) & label_mask) / num_p
  gmr_var = (num_n_above / num_n[:, None]).var(
      axis=-1, where=(labels > 0) & label_mask
  )
  return gmr, gmr_var



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mass-embedding functionality."""

import numpy as np
from perch_hoplite.zoo import model_configs
from perch_hoplite.zoo import zoo_interface

from absl.testing import absltest
from absl.testing import parameterized


class ZooTest(parameterized.TestCase):

  def test_pooled_embeddings(self):
    outputs = zoo_interface.InferenceOutputs(
        embeddings=np.zeros([10, 2, 8]), batched=False
    )
    batched_outputs = zoo_interface.InferenceOutputs(
        embeddings=np.zeros([3, 10, 2, 8]), batched=True
    )

    # Check that no-op is no-op.
    non_pooled = outputs.pooled_embeddings('', '')
    self.assertSequenceEqual(non_pooled.shape, outputs.embeddings.shape)
    batched_non_pooled = batched_outputs.pooled_embeddings('', '')
    self.assertSequenceEqual(
        batched_non_pooled.shape, batched_outputs.embeddings.shape
    )

    for pooling_method in zoo_interface.POOLING_METHODS:
      if pooling_method == 'squeeze':
        # The 'squeeze' pooling method throws an exception if axis size is > 1.
        with self.assertRaises(ValueError):
          outputs.pooled_embeddings(pooling_method, '')
        continue
      elif pooling_method == 'flatten':
        # Concatenates over the target axis.
        time_pooled = outputs.pooled_embeddings(pooling_method, '')
        self.assertSequenceEqual(time_pooled.shape, [2, 80])
        continue

      time_pooled = outputs.pooled_embeddings(pooling_method, '')
      self.assertSequenceEqual(time_pooled.shape, [2, 8])
      batched_time_pooled = batched_outputs.pooled_embeddings(
          pooling_method, ''
      )
      self.assertSequenceEqual(batched_time_pooled.shape, [3, 2, 8])

      channel_pooled = outputs.pooled_embeddings('', pooling_method)
      self.assertSequenceEqual(channel_pooled.shape, [10, 8])
      batched_channel_pooled = batched_outputs.pooled_embeddings(
          '', pooling_method
      )
      self.assertSequenceEqual(batched_channel_pooled.shape, [3, 10, 8])

      both_pooled = outputs.pooled_embeddings(pooling_method, pooling_method)
      self.assertSequenceEqual(both_pooled.shape, [8])
      batched_both_pooled = batched_outputs.pooled_embeddings(
          pooling_method, pooling_method
      )
      self.assertSequenceEqual(batched_both_pooled.shape, [3, 8])

  def test_simple_model_configs(self):
    """Load check for configs without framework dependencies."""
    for model_config_name in [
        model_configs.ModelConfigName.PLACEHOLDER,
        model_configs.ModelConfigName.BEANS_BASELINE,
    ]:
      with self.subTest(model_config_name):
        preset_info = model_configs.get_preset_model_config(model_config_name)
        self.assertGreaterEqual(preset_info.embedding_dim, 0)

  def test_beans_baseline_model(self):
    """Load check for configs with framework dependencies."""
    model = model_configs.load_model_by_name(
        model_configs.ModelConfigName.BEANS_BASELINE
    )
    fake_audio = np.zeros([5 * 32000], dtype=np.float32)
    outputs = model.embed(fake_audio)
    self.assertSequenceEqual(outputs.embeddings.shape, [5, 1, 80])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the embeddings databases interface."""

from perch_hoplite.db import datatypes as interface

from absl.testing import absltest


class DynamicInfoTest(absltest.TestCase):

  def test_dynamic_attributes(self):
    info = interface.DynamicInfo(
        foo="bar",
        bar="baz",
        value=100,
    )
    self.assertEqual(info.foo, "bar")
    self.assertEqual(info.bar, "baz")
    self.assertEqual(info.value, 100)
    self.assertRaises(AttributeError, lambda: info.status)
    info.status = "active"
    self.assertEqual(info.status, "active")

  def test_repr(self):
    info = interface.DynamicInfo(
        foo="bar",
        bar="baz",
        value=100,
    )
    self.assertEqual(
        repr(info),
        "DynamicInfo(foo='bar', bar='baz', value=100)",
    )
    info.status = "active"
    self.assertEqual(
        repr(info),
        "DynamicInfo(foo='bar', bar='baz', value=100, status='active')",
    )

  def test_eq(self):
    info1 = interface.DynamicInfo(
        foo="bar",
        bar="baz",
        value=100,
    )

    info2 = interface.DynamicInfo(
        foo="bar",
        bar="baz",
        value=100,
    )
    self.assertEqual(info1, info2)

    info3 = interface.DynamicInfo(
        foo="bar",
        bar="baz",
        value=101,
    )
    self.assertNotEqual(info1, info3)

    info4 = interface.DynamicInfo(
        foo="bar",
        bar="baz",
        value=100,
        status="active",
    )
    self.assertNotEqual(info1, info4)

    self.assertNotEqual(info1, "some string")


class DeploymentTest(absltest.TestCase):

  def test_required_attributes(self):
    dep = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
    )
    self.assertEqual(dep.id, 101)
    self.assertEqual(dep.name, "Sensor-A")
    self.assertEqual(dep.project, "River Monitoring")
    self.assertIsNone(dep.latitude)
    self.assertIsNone(dep.longitude)

  def test_optional_attributes(self):
    dep = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
        latitude=51.51,
        longitude=0.0,
    )
    self.assertAlmostEqual(dep.latitude, 51.51)
    self.assertAlmostEqual(dep.longitude, 0.0)

  def test_dynamic_attributes(self):
    dep = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
        latitude=51.51,
        status="active",
    )
    self.assertEqual(dep.status, "active")
    self.assertRaises(AttributeError, lambda: dep.country)
    dep.country = "United Kingdom"
    self.assertEqual(dep.country, "United Kingdom")

  def test_missing_required_argument(self):
    self.assertRaises(TypeError, lambda: interface.Deployment(name="Sensor-A"))
    self.assertRaises(
        TypeError, lambda: interface.Deployment(id=101, name="Sensor-A")
    )

  def test_repr(self):
    dep = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
    )
    self.assertEqual(
        repr(dep),
        "Deployment(id=101, name='Sensor-A', project='River Monitoring',"
        " latitude=None, longitude=None)",
    )
    dep.latitude = 51.51
    dep.country = "United Kingdom"
    self.assertEqual(
        repr(dep),
        "Deployment(id=101, name='Sensor-A', project='River Monitoring',"
        " latitude=51.51, longitude=None, country='United Kingdom')",
    )

  def test_eq(self):
    dep1 = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
    )

    dep2 = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
    )
    self.assertEqual(dep1, dep2)

    dep3 = interface.Deployment(
        id=102,
        name="Sensor-B",
        project="River Monitoring",
    )
    self.assertNotEqual(dep1, dep3)

    dep4 = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
        latitude=51.51,
    )
    self.assertNotEqual(dep1, dep4)

    dep5 = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
        status="active",
    )
    self.assertNotEqual(dep1, dep5)

    dep6 = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
        status="active",
    )
    self.assertEqual(dep5, dep6)

    dep7 = interface.Deployment(
        id=101,
        name="Sensor-A",
        project="River Monitoring",
        status="inactive",
    )
    self.assertNotEqual(dep6, dep7)

    self.assertNotEqual(dep1, "some string")


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convenience functions for predefined model configs."""

import dataclasses
import enum
import importlib

from ml_collections import config_dict
from perch_hoplite.zoo import hub
from perch_hoplite.zoo import zoo_interface


class ModelConfigName(enum.Enum):
  """Names of known preset configs."""

  BIRDNET_V2_1 = 'birdnet_V2.1'
  BIRDNET_V2_2 = 'birdnet_V2.2'
  BIRDNET_V2_3 = 'birdnet_V2.3'
  PERCH_V2 = 'perch_v2'
  PERCH_V2_GPU = 'perch_v2_gpu'
  PERCH_V2_CPU = 'perch_v2_cpu'
  PERCH_8 = 'perch_8'
  SURFPERCH = 'surfperch'
  VGGISH = 'vggish'
  YAMNET = 'yamnet'
  HUMPBACK = 'humpback'
  MULTISPECIES_WHALE = 'multispecies_whale'
  BEANS_BASELINE = 'beans_baseline'
  AVES = 'aves'
  BIRDAVES = 'birdaves'
  PLACEHOLDER = 'placeholder'


@dataclasses.dataclass
class PresetInfo:
  """Metadata for loading a specific model.

  Attributes:
    preset_name: The name of the preset.
    model_config: The model config.
    model_key: The short name for the model class.
    embedding_dim: The embedding dimension of the model.
  """

  preset_name: str
  model_config: config_dict.ConfigDict
  model_key: str
  embedding_dim: int

  def get_model_class(self) -> type[zoo_interface.EmbeddingModel]:
    """Convenience method to get the model class for this preset."""
    return get_model_class(self.model_key)

  def load_model(self) -> zoo_interface.EmbeddingModel:
    """Loads the embedding model."""
    return get_model_class(self.model_key).from_config(self.model_config)


def has_gpu_tf() -> bool:
  """Returns True if the system has a GPU and TensorFlow installed."""
  try:
    import tensorflow as tf  # pylint: disable=g-import-not-at-top
  except ImportError:
    return False
  return bool(tf.test.gpu_device_name())


def _get_obj(module, attr_name):
  """Get a thing from a module by name."""
  if hasattr(module, attr_name):
    return getattr(module, attr_name)
  else:
    raise ValueError(f'Module {module} does not have attribute {attr_name}')


def get_model_class(model_key: str) -> type[zoo_interface.EmbeddingModel]:
  """Import and return the model class."""
  if model_key in (
      'taxonomy_model_tf',
      'perch_8',
      'perch_v2',
      'perch_v2_cpu',
      'surfperch',
  ):
    module = importlib.import_module('perch_hoplite.zoo.taxonomy_model_tf')
    return _get_obj(module, 'TaxonomyModelTF')
  elif model_key == 'google_whale':
    module = importlib.import_module('perch_hoplite.zoo.models_tf')
    return _get_obj(module, 'GoogleWhaleModel')
  elif model_key == 'placeholder_model':
    module = importlib.import_module('perch_hoplite.zoo.placeholder_model')
    return _get_obj(module, 'PlaceholderModel')
  elif model_key == 'birdnet':
    module = importlib.import_module('perch_hoplite.zoo.models_tf')
    return _get_obj(module, 'BirdNet')
  elif model_key == 'tfhub_model':
    module = importlib.import_module('perch_hoplite.zoo.models_tf')
    return _get_obj(module, 'TFHubModel')
  elif model_key == 'aves':
    module = importlib.import_module('perch_hoplite.zoo.aves_model')
    return _get_obj(module, 'AVES')
  elif model_key == 'handcrafted_features_model':
    module = importlib.import_module(
        'perch_hoplite.zoo.handcrafted_features_model'
    )
    return _get_obj(module, 'HandcraftedFeaturesModel')
  else:
    raise ValueError(f'Unknown model key: {model_key}')


def load_model_by_name(
    model_config_name: str | ModelConfigName,
) -> zoo_interface.EmbeddingModel:
  """Loads the embedding model by model name."""
  model_config_name = ModelConfigName(model_config_name)
  preset_info = get_preset_model_config(model_config_name)
  return preset_info.load_model()


def perch_config(
    model_path: str,
    tfhub_version: int | None = None,
) -> config_dict.ConfigDict:
  """Returns a config for a Perch model."""
  model_config = config_dict.ConfigDict()
  model_config.model_path = model_path
  model_config.tfhub_version = tfhub_version
  model_config.window_size_s = 5.0
  model_config.hop_size_s = 5.0
  model_config.sample_rate = 32000
  return model_config


def get_preset_model_config(preset_name: str | ModelConfigName) -> PresetInfo:
  """Get a config_dict for a known model."""
  model_config = config_dict.ConfigDict()
  preset_name = ModelConfigName(preset_name)

  if preset_name == ModelConfigName.PERCH_8:
    model_key = 'taxonomy_model_tf'
    embedding_dim = 1280
    model_config.window_size_s = 5.0
    model_config.hop_size_s = 5.0
    model_config.sample_rate = 32000
    model_config.tfhub_path = (
        'google/bird-vocalization-classifier/'
        'tensorFlow2/bird-vocalization-classifier'
    )
    model_config.tfhub_version = 8
    model_config.model_path = ''
  elif preset_name == ModelConfigName.PERCH_V2:
    # Check for GPU support, and load the appropriate model.
    if has_gpu_tf():
      return get_preset_model_config(ModelConfigName.PERCH_V2_GPU)
    else:
      return get_preset_model_config(ModelConfigName.PERCH_V2_CPU)
  elif preset_name == ModelConfigName.PERCH_V2_GPU:
    model_key = 'taxonomy_model_tf'
    embedding_dim = 1536
    model_config.window_size_s = 5.0
    model_config.hop_size_s = 5.0
    model_config.sample_rate = 32000
    taxonomy_model_tf = importlib.import_module(
        'perch_hoplite.zoo.taxonomy_model_tf')
    model_config.tfhub_path = hub.PERCH_V2_SLUG
    model_config.tfhub_version = 2
    model_config.model_path = ''
  elif preset_name == ModelConfigName.PERCH_V2_CPU:
    model_key = 'taxonomy_model_tf'
    embedding_dim = 1536
    model_config.window_size_s = 5.0
    model_config.hop_size_s = 5.0
    model_config.sample_rate = 32000
    taxonomy_model_tf = importlib.import_module(
        'perch_hoplite.zoo.taxonomy_model_tf')
    model_config.tfhub_path = hub.PERCH_V2_CPU_SLUG
    model_config.tfhub_version = 1
    model_config.model_path = ''
  elif preset_name == ModelConfigName.HUMPBACK:
    model_key = 'google_whale'
    embedding_dim = 2048
    model_config.window_size_s = 3.9124
    model_config.sample_rate = 10000
    model_config.model_url = 'google/humpback-whale/tensorFlow2/humpback-whale'
    model_config.peak_norm = 0.02
  elif preset_name == ModelConfigName.MULTISPECIES_WHALE:
    model_key = 'google_whale'
    embedding_dim = 1280
    model_config.window_size_s = 5.0  # Is this correct?
    model_config.sample_rate = 24000
    model_config.model_url = 'google/multispecies-whale/tensorFlow2/default'
    model_config.peak_norm = -1.0
  elif preset_name == ModelConfigName.SURFPERCH:
    model_key = 'taxonomy_model_tf'
    embedding_dim = 1280
    model_config.window_size_s = 5.0
    model_config.hop_size_s = 5.0
    model_config.sample_rate = 32000
    taxonomy_model_tf = importlib.import_module(
        'perch_hoplite.zoo.taxonomy_model_tf')
    model_config.tfhub_version = 1
    model_config.tfhub_path = hub.SURFPERCH_SLUG
    model_config.model_path = ''
  elif preset_name.value.startswith('birdnet'):
    model_key = 'birdnet'
    birdnet_version = preset_name.value.split('_')[-1]
    if birdnet_version not in ('V2.1', 'V2.2', 'V2.3'):
      raise ValueError(f'Birdnet version not supported: {birdnet_version}')
    base_path = 'gs://chirp-public-bucket/models/birdnet'
    if birdnet_version == 'V2.1':
      embedding_dim = 420
      model_path = 'V2.1/BirdNET_GLOBAL_2K_V2.1_Model_FP16.tflite'
    elif birdnet_version == 'V2.2':
      embedding_dim = 320
      model_path = 'V2.2/BirdNET_GLOBAL_3K_V2.2_Model_FP16.tflite'
    elif birdnet_version == 'V2.3':
      embedding_dim = 1024
      model_path = 'V2.3/BirdNET_GLOBAL_3K_V2.3_Model_FP16.tflite'
    else:
      # TODO(tomdenton): Support V2.4.
      raise ValueError(f'Birdnet version not supported: {birdnet_version}')
    model_config.window_size_s = 3.0
    model_config.hop_size_s = 3.0
    model_config.sample_rate = 48000
    model_config.model_path = f'{base_path}/{model_path}'
    # Note: The v2_1 class list is appropriate for Birdnet 2.1, 2.2, and 2.3.
    model_config.class_list_name = 'birdnet_v2_1'
    model_config.num_tflite_threads = 4
  elif preset_name == ModelConfigName.YAMNET:
    model_key = 'tfhub_model'
    embedding_dim = 1024
    model_config.sample_rate = 16000
    model_config.model_url = hub.YAMNET_SLUG
    model_config.embedding_index = 1
    model_config.logits_index = 0
  elif preset_name == ModelConfigName.VGGISH:
    model_key = 'tfhub_model'
    embedding_dim = 128
    model_config.sample_rate = 16000
    model_config.model_url = hub.VGGISH_SLUG
    model_config.embedding_index = -1
    model_config.logits_index = -1
  elif preset_name == ModelConfigName.AVES:
    model_key = 'aves'
    embedding_dim = 768
    model_config.sample_rate = 16000
    module = importlib.import_module('perch_hoplite.zoo.aves_model')
    cache_fn = _get_obj(module, 'cache_onnx_model')
    model_config.model_path = cache_fn(
       'https://storage.googleapis.com/esp-public-files/ported_aves/aves-base-bio.onnx'
    )
  elif preset_name == ModelConfigName.BIRDAVES:
    model_key = 'aves'
    embedding_dim = 768
    model_config.sample_rate = 16000
    module = importlib.import_module('perch_hoplite.zoo.aves_model')
    cache_fn = _get_obj(module, 'cache_onnx_model')
    model_config.model_path = cache_fn(
        'https://storage.googleapis.com/esp-public-files/birdaves/birdaves-biox-large.onnx'
    )
  elif preset_name == ModelConfigName.PLACEHOLDER:
    model_key = 'placeholder'
    embedding_dim = 128
    model_config.sample_rate = 16000
  elif preset_name == ModelConfigName.BEANS_BASELINE:
    model_key = 'handcrafted_features_model'
    module = importlib.import_module(
        'perch_hoplite.zoo.handcrafted_features_model'
    )
    model_class = _get_obj(module, 'HandcraftedFeaturesModel')
    model_config = model_class.beans_baseline_config()
    embedding_dim = 80
  else:
    raise ValueError('Unsupported model preset: %s' % preset_name)
  return PresetInfo(
      preset_name=preset_name.value,
      model_config=model_config,
      model_key=model_key,
      embedding_dim=embedding_dim,
  )



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Score functions for Hoplite."""

from typing import Protocol, overload, runtime_checkable
import numpy as np


@runtime_checkable
class ScoreFn(Protocol):
  """Protocol for score functions supporting single and batched inputs."""

  @overload
  def __call__(self, data: np.ndarray, query: np.ndarray) -> float:
    ...

  @overload
  def __call__(self, data: np.ndarray, query: np.ndarray) -> np.ndarray:
    ...

  def __call__(self, data: np.ndarray, query: np.ndarray) -> float | np.ndarray:
    ...


def get_score_fn(
    name: str, bias: float | None = None, target_score: float | None = None
) -> ScoreFn:
  """Get a score function by name."""
  if name == 'dot':
    score_fn = numpy_dot
  elif name == 'cos':
    score_fn = numpy_cos
  elif name == 'neg_euclidean':
    score_fn = numpy_neg_euclidean
  else:
    raise ValueError('Unknown score function: ', name)

  if bias is not None:
    bias_fn = lambda x, y: score_fn(x, y) + bias
  else:
    bias_fn = score_fn

  if target_score is not None:
    # We want 'up is good', so take the negative absolute value.
    targeted_fn = lambda x, y: -np.abs(bias_fn(x, y) - target_score)
  else:
    targeted_fn = bias_fn

  return targeted_fn


def numpy_dot(data: np.ndarray, query: np.ndarray) -> float | np.ndarray:
  """Simple numpy dot product, which allows for multiple queries."""
  if len(query.shape) > 1:
    # Tensordot is a little faster than dot for multiple queries, but slower
    # than dot for single queries.
    return np.tensordot(data, query, axes=(-1, -1))
  return np.dot(data, query)


def numpy_cos(data: np.ndarray, query: np.ndarray) -> float | np.ndarray:
  """Simple numpy cosine similarity, allowing multiple queries."""
  data_norms = np.linalg.norm(data, axis=-1, keepdims=True)
  query_norms = np.linalg.norm(query, axis=-1, keepdims=True)
  unit_data = data / data_norms
  unit_query = query / query_norms
  if len(query.shape) > 1:
    return np.tensordot(unit_data, unit_query, axes=(-1, -1))
  return np.dot(unit_data, unit_query)


def numpy_neg_euclidean(
    data: np.ndarray, query: np.ndarray
) -> float | np.ndarray:
  """Negative L2 distance allowing multiple queries."""
  data_norms = np.linalg.norm(data, axis=-1)
  if len(query.shape) > 1:
    query_norms = np.linalg.norm(query, axis=-1)
    dot_products = np.tensordot(data, query, axes=(-1, -1))
    pairs = data_norms[:, np.newaxis] + query_norms[np.newaxis, :]
    return -pairs + 2 * dot_products

  query_norm = np.linalg.norm(query)
  dot_products = np.dot(data, query)
  return -data_norms + 2 * dot_products + query_norm


def get_jax_dot():
  """Create a composite jax dot product function."""
  # Jax is an optional dependency.
  # pylint: disable-next=g-import-not-at-top
  from jax import numpy as jnp

  def composite_dot(data, query):
    if len(query.shape) > 1:
      return jnp.tensordot(data, query, axes=(-1, -1))
    else:
      # Numpy is faster for simple matrix-vector products.
      return np.dot(data, query)

  return composite_dot



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementations of inference interfaces for applying trained models."""

import dataclasses
import tempfile
from typing import Any

from absl import logging
from etils import epath
from ml_collections import config_dict
import numpy as np
from perch_hoplite.taxonomy import namespace
from perch_hoplite.taxonomy import namespace_db
from perch_hoplite.zoo import hub
from perch_hoplite.zoo import taxonomy_model_tf
from perch_hoplite.zoo import zoo_interface
import tensorflow as tf
import tensorflow.compat.v1 as tf1


@dataclasses.dataclass
class SeparateEmbedModel(zoo_interface.EmbeddingModel):
  """Wrapper for separate separation and embedding models.

  Note: Use the separation model's sample rate. The embedding model's sample
  rate is used to resample prior to computing the embedding.

  Attributes:
    separation_model: SeparationModelTF.
    embedding_model: TaxonomyModelTF.
    embed_raw: If True, the outputs will include embeddings of the original
      audio in addition to embeddings for the separated channels. The embeddings
      will have shape [T, C+1, D], with the raw audio embedding on channel 0.
  """

  separator_model_tf_config: config_dict.ConfigDict
  taxonomy_model_tf_config: config_dict.ConfigDict
  separation_model: 'SeparatorModelTF'
  embedding_model: taxonomy_model_tf.TaxonomyModelTF
  embed_raw: bool = True

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'SeparateEmbedModel':
    separation_model = SeparatorModelTF.from_config(
        config.separator_model_tf_config
    )
    embedding_model = taxonomy_model_tf.TaxonomyModelTF.from_config(
        config.taxonomy_model_tf_config
    )
    return cls(
        separation_model=separation_model,
        embedding_model=embedding_model,
        **config,
    )

  def __post_init__(self):
    if self.separation_model.sample_rate != self.embedding_model.sample_rate:
      raise ValueError(
          'Separation and embedding models must have matching rates.'
      )

  def embed(self, audio_array: np.ndarray) -> zoo_interface.InferenceOutputs:
    # Frame the audio according to the embedding model's config.
    # We then apply separation to each frame independently, and embed
    # the separated audio.
    framed_audio = self.frame_audio(
        audio_array,
        self.embedding_model.window_size_s,
        self.embedding_model.hop_size_s,
    )
    # framed_audio has shape [Frames, Time]
    separation_outputs = self.separation_model.batch_embed(framed_audio)
    # separated_audio has shape [F, C, T]
    separated_audio = separation_outputs.separated_audio
    if separated_audio is None:
      raise RuntimeError('Separation model returned None for separated audio.')

    if self.embed_raw:
      separated_audio = np.concatenate(
          [
              framed_audio[:, np.newaxis, : separated_audio.shape[-1]],
              separated_audio,
          ],
          axis=1,
      )
    num_frames = separated_audio.shape[0]
    num_channels = separated_audio.shape[1]
    num_samples = separated_audio.shape[2]
    separated_audio = np.reshape(separated_audio, [-1, num_samples])

    embedding_outputs = self.embedding_model.batch_embed(separated_audio)

    if embedding_outputs.embeddings is not None:
      # Batch embeddings have shape [Batch, Time, Channels, Features]
      # Time is 1 because we have framed using the embedding model's
      # window_size. The batch size is num_frames * num_channels.
      embedding_outputs.embeddings = np.reshape(
          embedding_outputs.embeddings, [num_frames, num_channels, -1]
      )

    # Take the maximum logits over the channels dimension.
    if embedding_outputs.logits is not None:
      max_logits = {}
      for k, v in embedding_outputs.logits.items():
        v = v.reshape([num_frames, num_channels, -1])
        max_logits[k] = np.max(v, axis=1)
    else:
      max_logits = None

    return zoo_interface.InferenceOutputs(
        embeddings=embedding_outputs.embeddings,
        logits=max_logits,
        # Because the separated audio is framed, it does not match the
        # outputs interface, so we do not return it.
        separated_audio=None,
    )


@dataclasses.dataclass
class BirbSepModelTF1(zoo_interface.EmbeddingModel):
  """Separation model from the Bird MixIT paper.

  Example usage:
  ```
  from perch_hoplite.zoo import models
  birbsep1_config = config_dict.ConfigDict({
    'model_path': $MODEL_PATH,
    'window_size_s': 60.0,
    'keep_raw_channel': False,
    'sample_rate': 22050,
  })
  birbsep1 = models.BirbSepModelTF1.from_config(birbsep1_config)
  outputs = birbsep1.embed($SOME_AUDIO)
  ```
  """

  model_path: str
  window_size_s: float
  keep_raw_channel: bool
  session: Any
  input_placeholder_ns: Any
  output_tensor_ns: Any

  @classmethod
  def _find_checkpoint(cls, model_path: str) -> str:
    # Publicly released model does not have a checkpoints directory file.
    ckpt = None
    for ckpt in sorted(
        tuple(epath.Path(model_path).glob('model.ckpt-*.index'))
    ):
      ckpt = ckpt.as_posix()[: -len('.index')]
    if ckpt is None:
      raise FileNotFoundError('Could not find checkpoint file.')
    return ckpt

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'BirbSepModelTF1':
    """Load model files and create TF1 session graph."""
    metagraph_path_ns = epath.Path(config.model_path) / 'inference.meta'
    checkpoint_path = cls._find_checkpoint(config.model_path)
    graph_ns = tf.Graph()
    sess_ns = tf1.Session(graph=graph_ns)
    with graph_ns.as_default():
      new_saver = tf1.train.import_meta_graph(metagraph_path_ns)
      new_saver.restore(sess_ns, checkpoint_path)
      input_placeholder_ns = graph_ns.get_tensor_by_name(
          'input_audio/receiver_audio:0'
      )
      output_tensor_ns = graph_ns.get_tensor_by_name('denoised_waveforms:0')
    session = sess_ns
    return cls(
        session=session,
        input_placeholder_ns=input_placeholder_ns,
        output_tensor_ns=output_tensor_ns,
        **config,
    )

  def embed(self, audio_array: Any) -> zoo_interface.InferenceOutputs:
    start_sample = 0
    window_size = int(self.window_size_s * self.sample_rate)
    sep_chunks = []
    raw_chunks = []
    while start_sample <= audio_array.shape[0]:
      audio_chunk = audio_array[start_sample : start_sample + window_size]
      raw_chunks.append(audio_chunk)
      separated_audio = self.session.run(
          self.output_tensor_ns,
          feed_dict={
              self.input_placeholder_ns: audio_chunk[np.newaxis, np.newaxis, :]
          },
      )
      # Drop the extraneous batch dimension.
      separated_audio = np.squeeze(separated_audio, axis=0)
      sep_chunks.append(separated_audio)
      start_sample += window_size

    raw_chunks = np.concatenate(raw_chunks, axis=0)
    sep_chunks = np.concatenate(sep_chunks, axis=-1)
    if self.keep_raw_channel:
      sep_chunks = np.concatenate(
          [sep_chunks, raw_chunks[np.newaxis, :]], axis=0
      )
    return zoo_interface.InferenceOutputs(separated_audio=sep_chunks)

  def batch_embed(
      self, audio_batch: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    return zoo_interface.batch_embed_from_embed_fn(self.embed, audio_batch)


@dataclasses.dataclass
class SeparatorModelTF(zoo_interface.EmbeddingModel):
  """Separator SavedModel.

  Attributes:
    model_path: Path to model files.
    frame_size: Audio frame size for separation model.
    model: Loaded TF SavedModel.
    class_list: Loaded class_list for the model's output logits.
    windows_size_s: Window size for framing audio in samples. The audio will be
      chunked into frames of size window_size_s, which may help avoid memory
      blowouts. If None, will simply treat all audio as a single frame.
  """

  model_path: str
  frame_size: int
  model: Any
  class_list: namespace.ClassList
  window_size_s: float | None = None

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'SeparatorModelTF':
    logging.info('Loading taxonomy separator model...')
    model = tf.saved_model.load(epath.Path(config.model_path) / 'savedmodel')
    label_csv_path = epath.Path(config.model_path) / 'label.csv'
    with label_csv_path.open('r') as f:
      class_list = namespace.ClassList.from_csv(f)
    return cls(model=model, class_list=class_list, **config)

  def embed(self, audio_array: np.ndarray) -> zoo_interface.InferenceOutputs:
    # Drop samples to allow reshaping to frame_size
    excess_samples = audio_array.shape[0] % self.frame_size
    if excess_samples > 0:
      audio_array = audio_array[:-excess_samples]
    framed_audio = self.frame_audio(
        audio_array, self.window_size_s, self.window_size_s
    )
    framed_audio = np.reshape(
        framed_audio,
        [
            framed_audio.shape[0],
            framed_audio.shape[1] // self.frame_size,
            self.frame_size,
        ],
    )

    sep_audio, all_logits, all_embeddings = self.model.infer_tf(
        framed_audio[:1]
    )
    for window in framed_audio[1:]:
      separated, logits, embeddings = self.model.infer_tf(window[np.newaxis, :])
      sep_audio = np.concatenate([sep_audio, separated], axis=0)
      all_logits = np.concatenate([all_logits, logits], axis=0)
      all_embeddings = np.concatenate([all_embeddings, embeddings], axis=0)
    all_embeddings = all_embeddings[:, np.newaxis, :]

    # Recombine batch and time dimensions.
    sep_audio = np.reshape(sep_audio, [-1, sep_audio.shape[-1]])
    all_logits = np.reshape(all_logits, [-1, all_logits.shape[-1]])
    all_embeddings = np.reshape(all_embeddings, [-1, all_embeddings.shape[-1]])
    return zoo_interface.InferenceOutputs(
        all_embeddings, {'label': all_logits}, sep_audio
    )

  def batch_embed(
      self, audio_batch: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    return zoo_interface.batch_embed_from_embed_fn(self.embed, audio_batch)


@dataclasses.dataclass
class BirdNet(zoo_interface.EmbeddingModel):
  """Wrapper for BirdNet models.

  Attributes:
    model_path: Path to the saved model checkpoint or TFLite file.
    model: The TF SavedModel or TFLite interpreter.
    tflite: Whether the model is a TFLite model.
    class_list: The loaded class list.
    window_size_s: Window size for framing audio in samples.
    hop_size_s: Hop size for inference.
    num_tflite_threads: Number of threads to use with TFLite model.
    class_list_name: Name of the BirdNet class list.
  """

  model_path: str
  model: Any
  tflite: bool
  class_list: namespace.ClassList
  window_size_s: float = 3.0
  hop_size_s: float = 3.0
  num_tflite_threads: int = 16
  class_list_name: str = 'birdnet_v2_1'

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'BirdNet':
    logging.info('Loading BirdNet model...')
    if config.model_path.endswith('.tflite'):
      tflite = True
      with tempfile.NamedTemporaryFile() as tmpf:
        model_file = epath.Path(config.model_path)
        model_file.copy(tmpf.name, overwrite=True)
        model = tf.lite.Interpreter(
            tmpf.name,
            num_threads=config.num_tflite_threads,
            experimental_op_resolver_type=tf.lite.experimental.OpResolverType.BUILTIN_WITHOUT_DEFAULT_DELEGATES,
        )
      model.allocate_tensors()
    else:
      tflite = False
      model = tf.saved_model.load(config.model_path)
    db = namespace_db.load_db()
    class_list = db.class_lists[config.class_list_name]
    return cls(
        model=model,
        tflite=tflite,
        class_list=class_list,
        **config,
    )

  def embed_saved_model(
      self, audio_array: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    """Get logits using the BirdNet SavedModel."""
    # Note that there is no easy way to get the embedding from the SavedModel.
    all_logits = self.model(audio_array[:1])
    for window in audio_array[1:]:
      logits = self.model(window[np.newaxis, :])
      all_logits = np.concatenate([all_logits, logits], axis=0)
    return zoo_interface.InferenceOutputs(
        None, {self.class_list_name: all_logits}, None
    )

  def embed_tflite(
      self, audio_array: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    """Create an embedding and logits using the BirdNet TFLite model."""
    input_details = self.model.get_input_details()[0]
    output_details = self.model.get_output_details()[0]
    embedding_idx = output_details['index'] - 1
    embeddings = []
    logits = []
    for audio in audio_array:
      self.model.set_tensor(
          input_details['index'], np.float32(audio)[np.newaxis, :]
      )
      self.model.invoke()
      logits.append(self.model.get_tensor(output_details['index']))
      embeddings.append(self.model.get_tensor(embedding_idx))
    # Create [Batch, 1, Features]
    embeddings = np.array(embeddings)
    logits = np.array(logits)
    return zoo_interface.InferenceOutputs(
        embeddings, {self.class_list_name: logits}, None
    )

  def embed(self, audio_array: np.ndarray) -> zoo_interface.InferenceOutputs:
    framed_audio = self.frame_audio(
        audio_array, self.window_size_s, self.hop_size_s
    )
    if self.tflite:
      return self.embed_tflite(framed_audio)
    else:
      return self.embed_saved_model(framed_audio)

  def batch_embed(
      self, audio_batch: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    return zoo_interface.batch_embed_from_embed_fn(self.embed, audio_batch)


@dataclasses.dataclass
class GoogleWhaleModel(zoo_interface.EmbeddingModel):
  """Wrapper for Google Humpback model.

  Attributes:
    model_url: Path to model files.
    window_size_s: Window size for framing audio in seconds. 3.915 seems to work
      well for the Humpback model.
    hop_size_s: Hop size for inference.
    peak_norm: Peak normalizaiton target. Ignore if <= 0.
    model: Loaded TF SavedModel.
  """

  model_url: str
  window_size_s: float
  peak_norm: float
  class_list: namespace.ClassList
  model: Any  # TF SavedModel

  @classmethod
  def load_humpback_model(
      cls,
      model_url: str = hub.HUMPBACK_SLUG,
      **kwargs,
  ) -> 'GoogleWhaleModel':
    model = hub.load(model_url)
    class_list = namespace.ClassList('humpback', ('humpback',))
    sample_rate = model.metadata()['input_sample_rate'].numpy()
    window_size_s = model.metadata()['context_width_samples'] / sample_rate
    return cls(
        model=model,
        sample_rate=sample_rate,
        window_size_s=window_size_s,
        model_url=model_url,
        peak_norm=0.02,
        class_list=class_list,
        **kwargs,
    )

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'GoogleWhaleModel':
    model = hub.load(config.model_url)
    class_names = tuple(
        [str(c.numpy(), 'utf8') for c in model.metadata()['class_names']]
    )
    class_list = namespace.ClassList('multispecies_whale', class_names)
    return cls(model=model, class_list=class_list, **config)

  @property
  def hop_size_s(self) -> float:
    return self.window_size_s

  def batch_embed(
      self, audio_batch: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    # Renormalize to 0.02 peak.
    if self.peak_norm > 0:
      audio_batch = self.normalize_audio(audio_batch, self.peak_norm)
    spectrogram = self.model.front_end(audio_batch[:, :, np.newaxis])
    framed_spec = tf.signal.frame(
        spectrogram, frame_length=128, frame_step=128, pad_end=True, axis=1
    )
    rebatched_spec = tf.reshape(framed_spec, [-1, 128, framed_spec.shape[-1]])
    embeddings = self.model.features(rebatched_spec)

    # Recover batch and time dimensions, and add a channel dimension.
    embeddings = tf.reshape(
        embeddings,
        [framed_spec.shape[0], framed_spec.shape[1], 1, embeddings.shape[-1]],
    )
    logits = self.model.logits(rebatched_spec)
    logits = tf.reshape(
        logits, [framed_spec.shape[0], framed_spec.shape[1], logits.shape[-1]]
    )
    logits = {self.class_list.namespace: logits}
    outputs = zoo_interface.InferenceOutputs(
        embeddings=embeddings,
        logits=logits,
        frontend=spectrogram,
        batched=True,
    )
    return outputs

  def embed(self, audio_array: np.ndarray) -> zoo_interface.InferenceOutputs:
    return zoo_interface.embed_from_batch_embed_fn(
        self.batch_embed, audio_array
    )


@dataclasses.dataclass
class TFHubModel(zoo_interface.EmbeddingModel):
  """Generic wrapper for TFHub models which produce embeddings."""

  model: Any  # TFHub loaded model.
  model_url: str
  embedding_index: int
  logits_index: int = -1

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'TFHubModel':
    model = hub.load(config.model_url)
    return cls(
        model=model,
        **config,
    )

  @classmethod
  def yamnet(cls):
    # Parent class takes a sample_rate arg which pylint doesn't find.
    config = config_dict.ConfigDict({
        'sample_rate': 16000,
        'model_url': 'google/yamnet/tensorFlow2/yamnet',
        'embedding_index': 1,
        'logits_index': 0,
    })
    return TFHubModel.from_config(config)

  @classmethod
  def vggish(cls):
    config = config_dict.ConfigDict({
        'sample_rate': 16000,
        'model_url': 'google/vggish/tensorFlow2/vggish',
        'embedding_index': -1,
        'logits_index': -1,
    })
    return TFHubModel.from_config(config)

  def embed(
      self, audio_array: np.ndarray[Any, np.dtype[Any]]
  ) -> zoo_interface.InferenceOutputs:
    outputs = self.model(audio_array)
    if self.embedding_index < 0:
      embeddings = outputs
    else:
      embeddings = outputs[self.embedding_index]
    if len(embeddings.shape) == 1:
      embeddings = embeddings[np.newaxis, :]
    elif len(embeddings.shape) != 2:
      raise ValueError('Embeddings should have shape [Depth] or [Time, Depth].')

    if self.logits_index >= 0:
      logits = {'label': outputs[self.logits_index]}
    else:
      logits = None
    embeddings = embeddings[:, np.newaxis, :]
    return zoo_interface.InferenceOutputs(embeddings, logits, None, False)

  def batch_embed(
      self, audio_batch: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    return zoo_interface.batch_embed_from_embed_fn(self.embed, audio_batch)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for testing agile modeling functionality."""

import os

import numpy as np
from scipy.io import wavfile


def make_wav_files(
    base_path, classes, filenames, file_len_s=1.0, sample_rate_hz=16000
):
  """Create a pile of wav files in a directory structure."""
  rng = np.random.default_rng(seed=42)
  for subdir in classes:
    subdir_path = os.path.join(base_path, subdir)
    os.mkdir(subdir_path)
    for filename in filenames:
      with open(
          os.path.join(subdir_path, f'{filename}_{subdir}.wav'), 'wb'
      ) as f:
        noise = rng.normal(scale=0.2, size=int(file_len_s * sample_rate_hz))
        wavfile.write(f, sample_rate_hz, noise)
  audio_glob = os.path.join(base_path, '*/*.wav')
  return audio_glob



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for processing data for the Agile classifier."""

from collections.abc import Iterator, Sequence
import dataclasses
import itertools
from typing import Any

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface


@dataclasses.dataclass
class LabeledExample:
  """Example container class for model training and evaluation."""

  idx: np.ndarray
  embedding: np.ndarray
  multihot: np.ndarray
  is_labeled_mask: np.ndarray

  @property
  def is_batched(self) -> bool:
    return len(self.embedding.shape) > 1

  @classmethod
  def create_batched(
      cls, sequence: Sequence['LabeledExample']
  ) -> 'LabeledExample':
    return LabeledExample(
        idx=np.array([s.idx for s in sequence]),
        embedding=np.stack([s.embedding for s in sequence], axis=0),
        multihot=np.stack([s.multihot for s in sequence], axis=0),
        is_labeled_mask=np.stack([s.is_labeled_mask for s in sequence], axis=0),
    )

  def join_batches(self, other: 'LabeledExample') -> 'LabeledExample':
    if not self.is_batched or not other.is_batched:
      raise ValueError('Both examples must be batched.')
    return LabeledExample(
        idx=np.concatenate([self.idx, other.idx]),
        embedding=np.concatenate([self.embedding, other.embedding]),
        multihot=np.concatenate([self.multihot, other.multihot]),
        is_labeled_mask=np.concatenate(
            [self.is_labeled_mask, other.is_labeled_mask]
        ),
    )


@dataclasses.dataclass
class DataManager:
  """Base class for managing data for training and evaluation."""

  target_labels: tuple[str, ...] | None
  db: interface.HopliteDBInterface
  batch_size: int
  rng: np.random.Generator

  def get_train_test_split(self) -> tuple[np.ndarray, np.ndarray]:
    """Create a train/test split over all labels.

    Returns:
      Two numpy arrays contianing train and eval embedding ids, respectively.
    """
    raise NotImplementedError('get_train_test_split is not implemented.')

  def get_target_labels(self) -> tuple[str, ...]:
    if self.target_labels is None:
      return tuple(self.db.get_all_labels())
    return self.target_labels

  def get_multihot_labels(self, idx: int) -> tuple[np.ndarray, np.ndarray]:
    """Create the multihot label for one example."""
    window = self.db.get_window(idx)
    annotations = self.db.get_all_annotations(
        config_dict.create(
            eq=dict(recording_id=window.recording_id),
            approx=dict(offsets=window.offsets),
        )
    )
    target_labels = self.get_target_labels()
    lbl_idxes = {label: i for i, label in enumerate(target_labels)}
    pos = np.zeros(len(target_labels), dtype=np.float32)
    neg = np.zeros(len(target_labels), dtype=np.float32)
    for annotation in annotations:
      if annotation.label not in lbl_idxes:
        continue
      if annotation.label_type == datatypes.LabelType.POSITIVE:
        pos[lbl_idxes[annotation.label]] += 1.0
      elif annotation.label_type == datatypes.LabelType.NEGATIVE:
        neg[lbl_idxes[annotation.label]] += 1.0
    count = pos + neg
    mask = count > 0
    denom = np.maximum(count, 1.0)
    multihot = pos / denom
    return multihot, mask

  def labeled_example_iterator(
      self, ids: np.ndarray, repeat: bool = False
  ) -> Iterator[LabeledExample]:
    """Create an iterator for training a classifier for target_labels.

    Args:
      ids: The embedding IDs to iterate over.
      repeat: If True, repeat the iterator indefinitely.

    Yields:
      LabeledExample objects.
    """
    ids = ids.copy()
    self.rng.shuffle(ids)
    q = 0
    while True:
      x = ids[q]
      x_emb = self.db.get_embedding(x)
      x_multihot, x_is_labeled = self.get_multihot_labels(x)
      yield LabeledExample(x, x_emb, x_multihot, x_is_labeled)
      q += 1
      if q >= len(ids) and repeat:
        q = 0
        self.rng.shuffle(ids)
      elif q >= len(ids):
        break

  def batched_example_iterator(
      self,
      labeled_ids: np.ndarray,
      repeat: bool = False,
      **unused_kwargs,
  ) -> Iterator[LabeledExample]:
    """Labeled training data iterator with weak negatives."""
    example_iterator = self.labeled_example_iterator(labeled_ids, repeat=repeat)
    for ex_batch in batched(example_iterator, self.batch_size):
      yield LabeledExample.create_batched(ex_batch)


@dataclasses.dataclass
class AgileDataManager(DataManager):
  """Collects labeled data for training classifiers.

  Attributes:
    target_labels: The labels to use for training.
    db: The database to pull labeled examples from.
    train_ratio: The ratio of labeled examples to use for training.
    min_eval_examples: The minimum number of labeled examples to use for
      evaluation. Note that this is only enforced for positive examples.
    batch_size: The batch size for training.
    weak_negatives_batch_size: The batch size for weak negatives.
    rng: The random number generator to use.
  """
  train_ratio: float
  min_eval_examples: int
  weak_negatives_batch_size: int

  def get_single_label_train_test_split(
      self, label: str
  ) -> tuple[np.ndarray, np.ndarray]:
    """Create a train/test split for a single label."""
    pos_ids = self.db.match_window_ids(
        annotations_filter=config_dict.create(
            eq=dict(
                label=label,
                label_type=datatypes.LabelType.POSITIVE,
            )
        )
    )
    neg_ids = self.db.match_window_ids(
        annotations_filter=config_dict.create(
            eq=dict(
                label=label,
                label_type=datatypes.LabelType.NEGATIVE,
            )
        )
    )
    if not pos_ids:
      print('Warning: No positive examples for label: ', label)
      return np.array([], np.int64), np.array([], np.int64)
    all_ids = np.union1d(pos_ids, neg_ids)
    pos_ids = np.unique(pos_ids)
    neg_ids = np.unique(neg_ids)
    # Create a test/train split.
    # Strategy is to carefully create an eval set, and then define the train
    # set as the complement of the eval set amongst the labeled data.
    self.rng.shuffle(pos_ids)
    self.rng.shuffle(neg_ids)

    eval_ratio = 1.0 - self.train_ratio
    if self.min_eval_examples > pos_ids.shape[0]:
      raise ValueError(
          f'min_eval_examples ({self.min_eval_examples}) is greater than the'
          f' number of positive examples ({pos_ids.shape[0]}) for label'
          f' {label}.'
      )
    n_pos_eval = np.maximum(
        int(eval_ratio * pos_ids.shape[0]), self.min_eval_examples
    )
    pos_eval_ids = pos_ids[:n_pos_eval]
    n_neg_eval = np.maximum(
        int(eval_ratio * neg_ids.shape[0]), self.min_eval_examples
    )
    neg_eval_ids = neg_ids[:n_neg_eval]

    eval_ids = np.concatenate([pos_eval_ids, neg_eval_ids], axis=0)
    # De-dupe the eval_ids, in case there are multiple Labels for the embedding.
    eval_ids = np.unique(eval_ids)
    train_ids = np.setdiff1d(all_ids, eval_ids)
    return train_ids, eval_ids

  def get_train_test_split(self) -> tuple[np.ndarray, np.ndarray]:
    """Create a train/test split over all labels.

    Returns:
      Two numpy arrays contianing train and eval embedding ids, respectively.
    """
    train_ids, eval_ids = [], []
    for label in self.get_target_labels():
      lbl_train, lbl_eval = self.get_single_label_train_test_split(label)
      train_ids.append(lbl_train)
      eval_ids.append(lbl_eval)
    # We need to be careful not to allow labeled examples from any eval sets
    # to leak into the training set, which is a danger for multilabel examples.
    # Take the union of all eval sets, and use the complement as the train set.
    all_eval = np.concatenate(eval_ids, axis=0)
    all_eval = np.unique(all_eval)
    all_ids = np.concatenate(train_ids + eval_ids, axis=0)
    all_ids = np.unique(all_ids)
    all_train = np.setdiff1d(all_ids, all_eval)
    return all_train, all_eval

  def batched_example_iterator(
      self,
      labeled_ids: np.ndarray,
      repeat: bool = False,
      add_weak_negatives: bool = False,
  ) -> Iterator[LabeledExample]:
    """Labeled training data iterator with weak negatives."""
    example_iterator = self.labeled_example_iterator(labeled_ids, repeat=repeat)
    example_iterator = batched(example_iterator, self.batch_size)
    if not add_weak_negatives:
      for ex_batch in example_iterator:
        yield LabeledExample.create_batched(ex_batch)
      return

    weak_ids = np.setdiff1d(self.db.match_window_ids(), labeled_ids)
    weak_iterator = self.labeled_example_iterator(weak_ids, repeat=True)
    weak_iterator = batched(weak_iterator, self.weak_negatives_batch_size)

    for ex_batch, weak_ex_batch in zip(example_iterator, weak_iterator):
      # Join the two batches.
      ex_batch = LabeledExample.create_batched(ex_batch)
      weak_ex_batch = LabeledExample.create_batched(weak_ex_batch)
      if add_weak_negatives:
        yield ex_batch.join_batches(weak_ex_batch)
      else:
        yield ex_batch


@dataclasses.dataclass
class FullyAnnotatedDataManager(DataManager):
  """A DataManager for fully-annotated datasets."""

  train_examples_per_class: int
  min_eval_examples: int
  add_unlabeled_train_examples: bool

  def get_train_test_split(self) -> tuple[np.ndarray, np.ndarray]:
    """Create a train/test split over the fully-annotated dataset."""
    pos_id_sets = {}
    eval_id_sets = {}
    for label in self.get_target_labels():
      pos_id_sets[label] = self.db.match_window_ids(
          annotations_filter=config_dict.create(
              eq=dict(
                  label=label,
                  label_type=datatypes.LabelType.POSITIVE,
              )
          )
      )
      self.rng.shuffle(pos_id_sets[label])
      eval_id_sets[label] = pos_id_sets[label][: self.min_eval_examples]
    all_eval_ids = np.concatenate(tuple(eval_id_sets.values()), axis=0)

    # Now produce train sets of the desired size,
    # avoiding the selected eval examples.
    train_id_sets = {}
    for label in self.get_target_labels():
      pos_set = np.setdiff1d(pos_id_sets[label], all_eval_ids)
      train_id_sets[label] = pos_set[: self.train_examples_per_class]
    if self.add_unlabeled_train_examples:
      unlabeled_ids = np.setdiff1d(
          self.db.match_window_ids(),
          np.concatenate(tuple(pos_id_sets.values()), axis=0),
      )
      np.setdiff1d(unlabeled_ids, all_eval_ids)
      train_id_sets['UNLABELED'] = unlabeled_ids[
          : self.train_examples_per_class
      ]

    # The final eval set is the complement of all selected training id's.
    all_train_ids = np.concatenate(tuple(train_id_sets.values()), axis=0)
    eval_ids = np.setdiff1d(
        self.db.match_window_ids(),
        all_train_ids,
    )
    return all_train_ids, eval_ids


def batched(iterable: Iterator[Any], n: int) -> Iterator[Any]:
  # TODO(tomdenton): Use itertools.batched in Python 3.12+
  # batched('ABCDEFG', 3) → ABC DEF G
  if n < 1:
    raise ValueError('n must be at least one')
  iterator = iter(iterable)
  while batch := tuple(itertools.islice(iterator, n)):
    yield batch



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Search results containers."""

import dataclasses
import heapq

from perch_hoplite.db import datatypes


@dataclasses.dataclass
class SearchResult:
  """Container for a single search result."""

  # Embedding window ID.
  window_id: int

  # Score used for sorting the result.
  sort_score: float

  # Label type used for display. If None, then the label type is unknown.
  display_label_type: datatypes.LabelType | None = None

  def __lt__(self, other):
    return self.sort_score < other.sort_score

  def __gt__(self, other):
    return self.sort_score > other.sort_score

  def __le__(self, other):
    return self.sort_score <= other.sort_score

  def __ge__(self, other):
    return self.sort_score >= other.sort_score


@dataclasses.dataclass
class TopKSearchResults:
  """Top-K search results."""

  top_k: int
  search_results: list[SearchResult] = dataclasses.field(default_factory=list)
  _ids: set[int] = dataclasses.field(default_factory=set)

  def __post_init__(self):
    heapq.heapify(self.search_results)
    self._ids = set(q.window_id for q in self.search_results)

  def __iter__(self):
    for q in sorted(self.search_results, reverse=True):
      yield q

  def update(self, search_result: SearchResult, force_insert=False):
    """Update Results with the new result."""
    if not force_insert and self.will_filter(
        search_result.window_id, search_result.sort_score
    ):
      return
    if len(self.search_results) >= self.top_k:
      popped = heapq.heappop(self.search_results).window_id
      self._ids.remove(popped)
    heapq.heappush(self.search_results, search_result)
    self._ids.add(search_result.window_id)

  @property
  def min_score(self) -> float:
    return self.search_results[0].sort_score

  def will_filter(self, idx: int, score: float) -> bool:
    """Check whether a score is relevant."""
    if idx in self._ids:
      return True
    if len(self.search_results) < self.top_k:
      # Add the result, regardless of score, until we have k results.
      return False
    return score < self.min_score



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for source_info."""

import shutil
import tempfile

from perch_hoplite.agile import source_info
from perch_hoplite.agile.tests import test_utils

from absl.testing import absltest


class SourceInfoTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  def test_audio_sources_iteration(self):
    classes = ['pos', 'neg']
    filenames = ['foo', 'bar', 'baz']
    test_utils.make_wav_files(self.tempdir, classes, filenames, file_len_s=6.0)
    audio_sources = source_info.AudioSources(
        audio_globs=(
            source_info.AudioSourceConfig(
                dataset_name='pos',
                base_path=self.tempdir,
                file_glob='pos/*.wav',
            ),
            source_info.AudioSourceConfig(
                dataset_name='neg',
                base_path=self.tempdir,
                file_glob='neg/*.wav',
            ),
        )
    )

    with self.subTest('no_sharding'):
      shard_ids = tuple(audio_sources.iterate_all_sources())
      self.assertLen(shard_ids, len(classes) * len(filenames))

    audio_sources.audio_globs[0].shard_len_s = 2.0
    audio_sources.audio_globs[1].shard_len_s = 2.0
    with self.subTest('max_shards_less_than_file_len'):
      audio_sources.audio_globs[0].max_shards_per_file = 2
      audio_sources.audio_globs[1].max_shards_per_file = 2
      shard_ids = tuple(audio_sources.iterate_all_sources())
      self.assertLen(shard_ids, 12)

    with self.subTest('max_shards_matches_file_len'):
      audio_sources.audio_globs[0].max_shards_per_file = 3
      audio_sources.audio_globs[1].max_shards_per_file = 3
      shard_ids = tuple(audio_sources.iterate_all_sources())
      self.assertLen(shard_ids, 18)

    with self.subTest('max_shards_larger_than_file_len'):
      audio_sources.audio_globs[0].max_shards_per_file = 100
      audio_sources.audio_globs[1].max_shards_per_file = 100
      shard_ids = tuple(
          audio_sources.iterate_all_sources(
          )
      )
      self.assertLen(shard_ids, 18)

    with self.subTest('sharded_no_max_shards'):
      audio_sources.audio_globs[0].max_shards_per_file = None
      audio_sources.audio_globs[1].max_shards_per_file = None

      shard_ids = tuple(audio_sources.iterate_all_sources())
      self.assertLen(shard_ids, 18)

  def test_audio_glob_compatibility(self):
    audio_glob_1 = source_info.AudioSourceConfig(
        dataset_name='pos',
        base_path='/foo',
        file_glob='*.wav',
        target_sample_rate_hz=16000,
    )
    with self.subTest('different_base_path'):
      glob_diff_base_path = source_info.AudioSourceConfig(
          dataset_name='pos',
          base_path='/bar',
          file_glob='*.wav',
          target_sample_rate_hz=16000,
      )
      self.assertTrue(audio_glob_1.is_compatible(glob_diff_base_path))

    with self.subTest('different_dataset_name'):
      glob_diff_dataset_name = source_info.AudioSourceConfig(
          dataset_name='neg',
          base_path='/foo',
          file_glob='*.wav',
          target_sample_rate_hz=16000,
      )
      self.assertFalse(audio_glob_1.is_compatible(glob_diff_dataset_name))

    with self.subTest('different_target_sample_rate_hz'):
      glob_diff_sample_rate = source_info.AudioSourceConfig(
          dataset_name='pos',
          base_path='/foo',
          file_glob='*.wav',
          target_sample_rate_hz=24000,
      )
      self.assertFalse(audio_glob_1.is_compatible(glob_diff_sample_rate))

    with self.subTest('different_min_audio_len_s'):
      glob_diff_min_audio_len_s = source_info.AudioSourceConfig(
          dataset_name='pos',
          base_path='/foo',
          file_glob='*.wav',
          target_sample_rate_hz=16000,
          min_audio_len_s=2.0,
      )
      self.assertFalse(audio_glob_1.is_compatible(glob_diff_min_audio_len_s))

  def test_audio_sources_merge_update(self):
    audio_sources_1 = source_info.AudioSources(
        audio_globs=(
            source_info.AudioSourceConfig(
                dataset_name='pos',
                base_path=self.tempdir,
                file_glob='pos/*.wav',
            ),
            source_info.AudioSourceConfig(
                dataset_name='neg',
                base_path=self.tempdir,
                file_glob='neg/*.wav',
            ),
        )
    )
    with self.subTest('no_overlap'):
      disjoint = source_info.AudioSources(
          audio_globs=(
              source_info.AudioSourceConfig(
                  dataset_name='qua',
                  base_path=self.tempdir,
                  file_glob='qua/*.wav',
              ),
              source_info.AudioSourceConfig(
                  dataset_name='huh',
                  base_path=self.tempdir,
                  file_glob='huh/*.wav',
              ),
          )
      )
      got = audio_sources_1.merge_update(disjoint)
      self.assertLen(got.audio_globs, 4)

    with self.subTest('update_base_path'):
      overlap = source_info.AudioSources(
          audio_globs=(
              source_info.AudioSourceConfig(
                  dataset_name='pos',
                  base_path='/other/basedir',
                  file_glob='pos/*.wav',
              ),
          )
      )
      got = audio_sources_1.merge_update(overlap)
      self.assertLen(got.audio_globs, 2)
      self.assertEqual(got.audio_globs[0].base_path, '/other/basedir')

    with self.subTest('update_incompatible_audio_glob'):
      incompatible = source_info.AudioSources(
          audio_globs=(
              source_info.AudioSourceConfig(
                  dataset_name='pos',
                  base_path=self.tempdir,
                  file_glob='pos/*.wav',
                  target_sample_rate_hz=24000,
              ),
          )
      )
      with self.assertRaises(ValueError):
        audio_sources_1.merge_update(incompatible)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Zoo models using Jax."""

from collections.abc import Callable, Sequence
import dataclasses
import functools
import tempfile
from typing import Any

from etils import epath
import jax
from jax import numpy as jnp
from jax import random
import jaxonnxruntime
from jaxonnxruntime.core import call_onnx
from jaxonnxruntime.core import handler
from jaxonnxruntime.core import onnx_node
from ml_collections.config_dict import config_dict
import numpy as np
import onnx
from perch_hoplite.zoo import zoo_interface


@handler.register_op('InstanceNormalization')
class InstanceNormalization(handler.Handler):
  """Implementation of the ONNX InstanceNormalization operator."""

  @classmethod
  def _prepare(
      cls, node: onnx_node.OnnxNode, inputs: Sequence[Any], onnx_jax_impl: Any
  ):
    node.attrs_dict['epsilon'] = node.attrs.get('epsilon', 1e-5)

  @classmethod
  def version_6(
      cls, node: onnx_node.OnnxNode, inputs: Sequence[Any]
  ) -> Callable[..., Any]:
    """ONNX version_6 InstanceNormalization op."""
    cls._prepare(node, inputs, onnx_instancenormalization)
    return onnx_instancenormalization


@functools.partial(jax.jit, static_argnames=('epsilon',))
def onnx_instancenormalization(*input_args, epsilon: float):
  """https://github.com/onnx/onnx/blob/v1.12.0/docs/Operators.md#InstanceNormalization for more details."""
  input_, scale, b = input_args

  dims_input = len(input_.shape)
  input_mean = jnp.mean(
      input_, axis=tuple(range(dims_input))[2:], keepdims=True
  )
  input_var = jnp.var(input_, axis=tuple(range(dims_input))[2:], keepdims=True)

  dim_ones = (1,) * (dims_input - 2)
  scale = scale.reshape(-1, *dim_ones)
  b = b.reshape(-1, *dim_ones)

  return (input_ - input_mean) / jnp.sqrt(input_var + epsilon) * scale + b


def cache_onnx_model(url: str) -> str:
  """Cache the ONNX model at a local path."""
  url = epath.Path(url)
  filename = url.name
  # Check for existing file first.
  tempdir = tempfile.gettempdir()
  cached_model_path = epath.Path(tempdir) / filename
  if cached_model_path.exists():
    return cached_model_path.as_posix()
  # Download the model and cache it.
  url.copy(cached_model_path)
  return cached_model_path.as_posix()


@dataclasses.dataclass
class AVES(zoo_interface.EmbeddingModel):
  """Wrapper around AVES ONNX model.

  This model was originally trained to take audio with a 16 kHz sample rate.

  Each time the model gets called with a new input shape, a new JAX function
  gets created and compiled, and all parameters get copied. This could be
  slow.
  """

  model_path: str = ''
  model: onnx.onnx_ml_pb2.ModelProto = dataclasses.field(init=False)
  input_shape: tuple[int, ...] | None = dataclasses.field(
      default=None, init=False
  )
  model_func: Callable[[list[np.ndarray[Any, Any]]], list[jax.Array]] | None = (
      dataclasses.field(default=None, init=False)
  )

  @classmethod
  def from_config(
      cls, model_config: config_dict.ConfigDict
  ) -> zoo_interface.EmbeddingModel:
    return cls(**model_config)

  def __post_init__(self):
    jaxonnxruntime.config.update(
        'jaxort_only_allow_initializers_as_static_args', False
    )
    self.model = onnx.load(self.model_path)

  def embed(
      self, audio_array: np.ndarray[Any, Any]
  ) -> zoo_interface.InferenceOutputs:
    return zoo_interface.embed_from_batch_embed_fn(
        self.batch_embed, audio_array
    )

  def batch_embed(
      self, audio_batch: np.ndarray[Any, Any]
  ) -> zoo_interface.InferenceOutputs:
    # Compile new function if necessary
    if audio_batch.shape != self.input_shape:
      key = random.PRNGKey(0)
      random_input = random.normal(key, audio_batch.shape, dtype=jnp.float32)
      model_func, model_params = call_onnx.call_onnx_model(
          self.model, [random_input]
      )
      self.input_shape = audio_batch.shape
      self.model_func = functools.partial(model_func, model_params)

    # Embed and add a single channel dimension
    embeddings = np.asarray(self.model_func([audio_batch])[0])
    embeddings = embeddings[:, :, np.newaxis, :]
    return zoo_interface.InferenceOutputs(embeddings=embeddings, batched=True)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for sqlite-backed Agile modeling notebooks."""

import dataclasses

from etils import epath
from ml_collections import config_dict
from perch_hoplite.agile import embed
from perch_hoplite.agile import source_info
from perch_hoplite.db import db_loader
from perch_hoplite.db import sqlite_usearch_impl
from perch_hoplite.zoo import model_configs


@dataclasses.dataclass
class AgileConfigs:
  """Container for the various configs used in the Agile notebooks."""

  # Config for the raw audio sources.
  audio_sources_config: source_info.AudioSources
  # Database config for the embeddings database.
  db_config: db_loader.DBConfig
  # Config for the embedding model.
  model_config: embed.ModelConfig

  def as_config_dict(self) -> config_dict.ConfigDict:
    """Returns the configs as a ConfigDict."""
    return config_dict.ConfigDict({
        'audio_sources_config': self.audio_sources_config.to_config_dict(),
        'db_config': self.db_config.to_config_dict(),
        'model_config': self.model_config.to_config_dict(),
    })


def load_configs(
    audio_sources: source_info.AudioSources,
    db_path: str | None = None,
    model_config_key: str = 'perch_v2',
    db_key: str = 'sqlite_usearch',
) -> AgileConfigs:
  """Load default configs for the notebook and return them as an AgileConfigs.

  Args:
    audio_sources: Mapping from dataset name to pairs of `(root directory, file
      glob)`.
    db_path: Location of the database.  If None, the database will be created in
      the same directory as the audio.
    model_config_key: Name of the embedding model to use.
    db_key: The type of database to use.

  Returns:
    AgileConfigs object with the loaded configs.
  """
  if db_path is None:
    if len(audio_sources.audio_globs) > 1:
      raise ValueError(
          'db_path must be specified when embedding multiple datasets.'
      )
    # Put the DB in the same directory as the audio.
    db_path = epath.Path(next(iter(audio_sources.audio_globs)).base_path)

  preset_info = model_configs.get_preset_model_config(model_config_key)
  db_model_config = embed.ModelConfig(
      model_key=preset_info.model_key,
      embedding_dim=preset_info.embedding_dim,
      model_config=preset_info.model_config,
  )
  db_config = config_dict.ConfigDict({
      'db_path': db_path,
  })
  if db_key == 'sqlite_usearch':
    # A sane default.
    db_config.usearch_cfg = sqlite_usearch_impl.get_default_usearch_config(
        preset_info.embedding_dim
    )

  return AgileConfigs(
      audio_sources_config=audio_sources,
      db_config=db_loader.DBConfig(db_key, db_config),
      model_config=db_model_config,
  )



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for testing."""

from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import datatypes
from perch_hoplite.db import in_mem_impl
from perch_hoplite.db import interface
from perch_hoplite.db import sqlite_usearch_impl

# DB types for testing.
DB_TYPES = ('in_mem', 'sqlite_usearch')
DB_TYPE_NAMED_PAIRS = (('in_mem-sqlite_usearch', 'in_mem', 'sqlite_usearch'),)
PERSISTENT_DB_TYPES = ('sqlite_usearch',)

CLASS_LABELS = ('alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta')


def make_db(
    path: str,
    db_type: str,
    num_embeddings: int,
    rng: np.random.Generator,
    embedding_dim: int = 128,
    fill_random: bool = True,
) -> interface.HopliteDBInterface:
  """Create a test DB of the specified type."""
  if db_type == 'in_mem':
    db = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=embedding_dim)
  elif db_type == 'sqlite_usearch':
    usearch_cfg = sqlite_usearch_impl.get_default_usearch_config(embedding_dim)
    db = sqlite_usearch_impl.SQLiteUSearchDB.create(
        db_path=path, usearch_cfg=usearch_cfg
    )
  else:
    raise ValueError(f'Unknown db type: {db_type}')
  # Insert a few embeddings...
  if fill_random:
    insert_random_embeddings(db, embedding_dim, num_embeddings, rng)
  config = config_dict.ConfigDict()
  config.embedding_dim = embedding_dim
  db.insert_metadata('db_config', config)
  model_config = config_dict.ConfigDict()
  model_config.embedding_dim = embedding_dim
  model_config.model_name = 'fake_model'
  db.insert_metadata('model_config', model_config)
  db.commit()
  return db


def insert_random_embeddings(
    db: interface.HopliteDBInterface,
    emb_dim: int = 1280,
    num_embeddings: int = 1000,
    seed: int = 42,
):
  """Insert randomly generated embedding vectors into the DB."""
  rng = np.random.default_rng(seed=seed)
  np_alpha = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789')

  projects = ('a', 'b', 'c')
  deployment_ids = [
      db.insert_deployment(name=f'deployment_{project}', project=project)
      for project in projects
  ]

  window_size_s = 5.0
  for _ in range(num_embeddings):
    deployment_id = rng.choice(deployment_ids).item()
    filename = ''.join(
        [str(a) for a in rng.choice(np_alpha, size=8, replace=False)]
    )
    recording_id = db.insert_recording(
        filename=filename, deployment_id=deployment_id
    )

    embedding = np.float32(rng.normal(size=emb_dim, loc=0, scale=0.1))
    offsets = rng.integers(0, 100, size=[1])
    offsets = [offsets[0], offsets[0] + window_size_s]
    db.insert_window(
        recording_id, offsets, embedding, handle_duplicates='allow'
    )
  db.commit()


def clone_embeddings(
    source_db: interface.HopliteDBInterface,
    target_db: interface.HopliteDBInterface,
):
  """Copy all embeddings to target_db and provide an id mapping."""

  # First, clone deployments and keep a map between source and target ids.
  deployment_id_mapping = {None: None}
  for deployment in source_db.get_all_deployments():
    target_id = target_db.insert_deployment(**deployment.to_kwargs(skip=['id']))
    deployment_id_mapping[deployment.id] = target_id

  # Second, clone recordings and keep a map between source and target ids.
  recording_id_mapping = {}
  for recording in source_db.get_all_recordings():
    target_id = target_db.insert_recording(
        deployment_id=deployment_id_mapping[recording.deployment_id],
        **recording.to_kwargs(skip=['id', 'deployment_id']),
    )
    recording_id_mapping[recording.id] = target_id

  # Finally, clone windows and keep a map between source and target ids.
  window_id_mapping = {}
  for window in source_db.get_all_windows():
    target_id = target_db.insert_window(
        recording_id=recording_id_mapping[window.recording_id],
        embedding=source_db.get_embedding(window.id),
        handle_duplicates='allow',
        **window.to_kwargs(skip=['id', 'embedding', 'recording_id']),
    )
    window_id_mapping[window.id] = target_id

  # Return the window id mapping.
  return window_id_mapping


def add_random_labels(
    db: interface.HopliteDBInterface,
    rng: np.random.Generator,
    unlabeled_prob: float = 0.5,
    positive_label_prob: float = 0.5,
    provenance: str = 'test',
):
  """Insert random labels for a subset of embeddings."""
  for idx in db.match_window_ids():
    if rng.random() < unlabeled_prob:
      continue
    if rng.random() < positive_label_prob:
      label_type = datatypes.LabelType.POSITIVE
    else:
      label_type = datatypes.LabelType.NEGATIVE
    window = db.get_window(idx)
    db.insert_annotation(
        recording_id=window.recording_id,
        offsets=window.offsets,
        label=str(rng.choice(CLASS_LABELS)),
        label_type=label_type,
        provenance=provenance,
        handle_duplicates='allow',
    )
  db.commit()
